import glob
import os
from configparser import NoSectionError, MissingSectionHeaderError, NoOptionError, ParsingError
from typing import Set

from rlbot.parsing.bot_config_bundle import BotConfigBundle, ScriptConfigBundle, get_bot_config_bundle, get_script_config_bundle


def scan_directory_for_bot_configs(root_dir) -> Set[BotConfigBundle]:
    """
    Recursively scans a directory for all valid bot configs.
    :param root_dir: Directory to scan.
    :return: The set of bot configs that were found.
    """

    configs = set()

    for filename in glob.iglob(os.path.join(root_dir, '**/*.cfg'), recursive=True):
        try:
            bundle = get_bot_config_bundle(filename)
            configs.add(bundle)
        except (NoSectionError, MissingSectionHeaderError, NoOptionError, AttributeError, ParsingError, FileNotFoundError) as ex:
            pass

    return configs

def scan_directory_for_script_configs(root_dir) -> Set[ScriptConfigBundle]:
    """
    Recursively scans a directory for all valid script configs.
    :param root_dir: Directory to scan.
    :return: The set of script configs that were found.
    """

    configs = set()

    for filename in glob.iglob(os.path.join(root_dir, '**/*.cfg'), recursive=True):
        try:
            bundle = get_script_config_bundle(filename)
            configs.add(bundle)
        except (NoSectionError, MissingSectionHeaderError, NoOptionError, AttributeError, ParsingError, FileNotFoundError):
            pass

    return configs



from typing import Union, Optional, Iterator, Iterable, Callable
import random
import time
import traceback
from contextlib import contextmanager

from rlbot.training.status_rendering import training_status_renderer_context, Row
from rlbot.matchconfig.match_config import MatchConfig
from rlbot.setup_manager import SetupManager
from rlbot.utils import rate_limiter
from rlbot.utils.game_state_util import GameState
from rlbot.utils.logging_utils import get_logger, DEFAULT_LOGGER
from rlbot.utils.rendering.rendering_manager import RenderingManager
from rlbot.utils.structures.game_data_struct import GameTickPacket
from rlbot.utils.structures.game_interface import GameInterface
from rlbot.matchcomms.client import MatchcommsClient

"""
This file contains a minimal API to implement training.
For a more useful API see: https://github.com/RLBot/RLBotTraining/
"""


# Extend Pass and/or Fail to add your own, more detailed metrics.

class Pass:
    """ Indicates that the bot passed the exercise. """

    def __repr__(self):
        return 'PASS'


class Fail:
    """ Indicates that the bot failed the exercise. """

    def __repr__(self):
        return 'FAIL'


class FailDueToExerciseException(Fail):
    """ Indicates that the test code threw an expetion. """

    def __init__(self, exception: Exception, traceback_string: str):
        self.exception = exception
        self.traceback_string = traceback_string

    def __repr__(self):
        return 'FAIL: Exception raised by Exercise:\n' + self.traceback_string


# Note: not using Grade as a abstract base class for Pass/Fail
#       as there should not be Grades which are neither Pass nor Fail.
Grade = Union[Pass, Fail]


class Exercise:
    """
    Statisfy this interface to define your test cases.
    This class provides a seeded random generator to support variation testing.
    The responsibility of detecting timeouts lies with the implementation of
    on_tick().
    """

    # Creates a matchcomms client connected to the current match.
    # Initialized before on_briefing() is called.
    _matchcomms_factory: Callable[[], MatchcommsClient] = None

    def get_name(self) -> str:
        """
        Gets the name to be displayed on screen.
        """
        raise NotImplementedError()

    def get_match_config(self) -> MatchConfig:
        """
        Gets the config with which this exercise should be run.
        """
        raise NotImplementedError()

    def setup(self, rng: random.Random) -> GameState:
        """
        Returns the state in which the game should start in.
        The implementing class is responsible for resetting the state after setup is called,
        such that the exercise can be run multiple times to get the same result.
        :param random: A seeded random number generator. For repeated runs of this
            exercise, this parameter and the bots should be the only things which
            causes variations between runs.
        """
        raise NotImplementedError()

    def on_tick(self, game_tick_packet: GameTickPacket) -> Optional[Grade]:
        """
        This method is called each tick to allow you to make an assessment of the
        performance of the bot(s).
        The order for whether on_tick comes before the bots recieving the packet is undefined.

        If this method returns None, the run of the exercise will continue.
        If this method returns Pass() or Fail() or raises an exceptions, the run of
        the exercise is terminated and any metrics will be returned.
        """
        raise NotImplementedError()

    def on_briefing(self) -> Optional[Grade]:
        """
        This method is called before state-setting such that bots can be "briefed" on the upcoming exercise.
        The "briefing" is usually for using matchcomms to convey objectives and parameters.
        A grade can be returned in case bot responded sufficient to pass or fail the exercise
        before any on_tick() grading happens.
        """
        pass

    def set_matchcomms_factory(self, matchcomms_factory: Callable[[], MatchcommsClient]):
        self._matchcomms_factory = matchcomms_factory

    def render(self, renderer: RenderingManager):
        """
        This method is called each tick to render exercise debug information.
        This method is called after on_tick().
        It is optional to override this method.
        """
        pass


class Result:
    def __init__(self, input_exercise: Exercise, input_seed: int, grade: Grade):
        assert grade
        self.seed = input_seed
        self.exercise = input_exercise
        self.grade = grade


def run_exercises(setup_manager: SetupManager, exercises: Iterable[Exercise], seed: int,
                  reload_agent: bool = True) -> Iterator[Result]:
    """
    It is recommended to use setup_manager_context() to generate your setup_manager.
    """
    game_interface = setup_manager.game_interface
    names = [exercise.get_name() for exercise in exercises]
    with training_status_renderer_context(names, game_interface.renderer) as ren:
        for i, exercise in enumerate(exercises):
            with safe_matchcomms_factory(setup_manager) as matchcomms_factory:
                def update_row(status: str, status_color_func):
                    nonlocal i
                    nonlocal exercise
                    ren.update(i, Row(exercise.get_name(), status, status_color_func))

                update_row('config', ren.renderman.white)
                # Only reload the match if the config has changed.
                new_match_config = exercise.get_match_config()
                if new_match_config != setup_manager.match_config:
                    update_row('match', ren.renderman.white)
                    _setup_match(new_match_config, setup_manager)
                    update_row('bots', ren.renderman.white)
                    _wait_until_bots_ready(setup_manager, new_match_config)

                if reload_agent:
                    update_row('reload', ren.renderman.white)
                    setup_manager.reload_all_agents(quiet=True)

                # Briefing
                update_row('brief', ren.renderman.white)
                try:
                    exercise.set_matchcomms_factory(matchcomms_factory)
                    early_result = exercise.on_briefing()
                except Exception as e:
                    update_row('brief', ren.renderman.red)
                    yield Result(exercise, seed, FailDueToExerciseException(e, traceback.format_exc()))
                    continue
                if early_result is not None:
                    if isinstance(early_result.grade, Pass):
                        update_row('PASS', ren.renderman.green)
                    else:
                        update_row('FAIL', ren.renderman.red)
                    yield early_result
                    continue

                update_row('wait', ren.renderman.white)
                _wait_until_good_ticks(game_interface)

                update_row('setup', ren.renderman.white)
                error_result = _setup_exercise(game_interface, exercise, seed)
                if error_result is not None:
                    update_row('setup', ren.renderman.red)
                    yield error_result
                    continue

                # Wait for the set_game_state() to propagate before we start running ex.on_tick()
                # TODO: wait until the game looks similar.
                update_row('sleep', ren.renderman.white)
                time.sleep(0.03)

                update_row('>>>>', ren.renderman.white)
                result = _grade_exercise(game_interface, exercise, seed)

                if isinstance(result.grade, Pass):
                    update_row('PASS', ren.renderman.green)
                else:
                    update_row('FAIL', ren.renderman.red)
                yield result

@contextmanager
def safe_matchcomms_factory(setup_manager: SetupManager):
    clients = []
    has_finished = False
    def matchcomms_factory() -> MatchcommsClient:
        assert not has_finished
        client = MatchcommsClient(setup_manager.matchcomms_server.root_url)
        clients.append(client)
        return client
    try:
        yield matchcomms_factory
    finally:
        has_finished = True
        for client in clients:
            client.close()



def _wait_until_bots_ready(setup_manager: SetupManager, match_config: MatchConfig):
    logger = get_logger(DEFAULT_LOGGER)
    while not setup_manager.has_received_metadata_from_all_bots():
        logger.debug('Waiting on all bots to post their metadata.')
        setup_manager.try_recieve_agent_metadata()
        time.sleep(0.1)


def _wait_until_good_ticks(game_interface: GameInterface, required_new_ticks: int=3):
    """Blocks until we're getting new packets, indicating that the match is ready."""
    rate_limit = rate_limiter.RateLimiter(120)
    last_tick_game_time = None  # What the tick time of the last observed tick was
    packet = GameTickPacket()  # We want to do a deep copy for game inputs so people don't mess with em
    seen_times = 0
    while seen_times < required_new_ticks:
        game_interface.update_live_data_packet(packet)
        def is_good_tick():
            if packet.game_info.seconds_elapsed == last_tick_game_time: return False
            if not packet.game_info.is_round_active: return False
            if any(car.is_demolished for car in packet.game_cars): return False
            return True
        if is_good_tick():
            seen_times += 1
        last_tick_game_time = packet.game_info.seconds_elapsed

        rate_limit.acquire()


def _setup_match(match_config: MatchConfig, manager: SetupManager):
    manager.shut_down(kill_all_pids=True, quiet=True)  # To be safe.
    manager.load_match_config(match_config)
    manager.launch_early_start_bot_processes()
    manager.start_match()
    manager.launch_bot_processes()


def _setup_exercise(game_interface: GameInterface, ex: Exercise, seed: int) -> Optional[Result]:
    """
    Set the game state.
    Only returns a result if there was an error in ex.setup()
    """
    rng = random.Random()
    rng.seed(seed)
    try:
        game_state = ex.setup(rng)
    except Exception as e:
        return Result(ex, seed, FailDueToExerciseException(e, traceback.format_exc()))
    game_interface.set_game_state(game_state)


def _grade_exercise(game_interface: GameInterface, ex: Exercise, seed: int) -> Result:
    grade = None
    rate_limit = rate_limiter.RateLimiter(120)
    last_tick_game_time = None  # What the tick time of the last observed tick was
    game_tick_packet = GameTickPacket()  # We want to do a deep copy for game inputs so people don't mess with em

    # Run until the Exercise finishes.
    while grade is None:

        # Read from game data shared memory
        game_interface.update_live_data_packet(game_tick_packet)

        # Run ex.on_tick() only if the game_info has updated.
        tick_game_time = game_tick_packet.game_info.seconds_elapsed
        if tick_game_time != last_tick_game_time:
            last_tick_game_time = tick_game_time
            try:
                grade = ex.on_tick(game_tick_packet)
                ex.render(game_interface.renderer)
            except Exception as e:
                return Result(ex, seed, FailDueToExerciseException(e, traceback.format_exc()))

        rate_limit.acquire()

    return Result(ex, seed, grade)



import json
from json import JSONDecodeError

from rlbot.matchconfig.loadout_config import LoadoutConfig, LoadoutPaintConfig, Color
from rlbot.parsing.custom_config import ConfigObject, ConfigHeader
from rlbot.utils.logging_utils import get_logger

PARTICIPANT_CONFIGURATION_HEADER = 'Participant Configuration'
PARTICIPANT_CONFIG_KEY = 'participant_config'
SCRIPT_CONFIGURATION_HEADER = 'Scripts'
SCRIPT_CONFIG_KEY = 'script_config'
PARTICIPANT_BOT_SKILL_KEY = 'participant_bot_skill'
PARTICIPANT_TYPE_KEY = 'participant_type'
PARTICIPANT_TEAM = 'participant_team'
PARTICIPANT_LOADOUT_CONFIG_KEY = 'participant_loadout_config'
BOT_CONFIG_LOADOUT_HEADER = 'Bot Loadout'
BOT_CONFIG_LOADOUT_ORANGE_HEADER = 'Bot Loadout Orange'
BOT_CONFIG_LOADOUT_PAINT_BLUE_HEADER = 'Bot Paint Blue'
BOT_CONFIG_LOADOUT_PAINT_ORANGE_HEADER = 'Bot Paint Orange'

logger = get_logger('rlbot')


def add_participant_header(config_object):
    participant_header = config_object.add_header_name(PARTICIPANT_CONFIGURATION_HEADER, is_indexed=True)
    participant_header.add_value(PARTICIPANT_CONFIG_KEY, str, default=None,
                                 description="""The location of the configuration file for your agent here.
                                             Only total_num_participants config files will be read!
                                             Everything needs a config, even players and default bots.
                                             We still set loadouts and names from config!""")

    participant_header.add_value(PARTICIPANT_TEAM, int, default=0,
                                 description="Which team the player should be on:" +
                                             "\nteam 0 (blue) shoots on positive goal, " +
                                             "team 1 (orange) shoots on negative goal")

    participant_header.add_value(PARTICIPANT_TYPE_KEY, str, default='rlbot',
                                 description="""Accepted values are "human", "rlbot", "psyonix" and "party_member_bot"
                                             You can have up to 4 local players and they must
                                             be activated in game or it will crash.
                                             If no player is specified you will be spawned in as spectator!
                                             human - not controlled by the framework
                                             rlbot - controlled by the framework
                                             psyonix - default bots (skill level can be changed with participant_bot_skill
                                             party_member_bot - controlled by an rlbot but the game detects it as a human""")

    participant_header.add_value(PARTICIPANT_BOT_SKILL_KEY, float, default=1.0,
                                 description='If participant is a bot and not RLBot controlled,' +
                                             ' this value will be used to set bot skill.\n' +
                                             '0.0 is Rookie, 0.5 is pro, 1.0 is all-star. ' +
                                             ' You can set values in-between as well.')

    participant_header.add_value(PARTICIPANT_LOADOUT_CONFIG_KEY, str, default="None",
                                 description="""A path to a loadout config file which will override the path in the agent config
                                             Use None to extract the path from the agent config""")


def add_scripts_header(config_object):
    participant_header = config_object.add_header_name(SCRIPT_CONFIGURATION_HEADER, is_indexed=True)
    participant_header.add_value(SCRIPT_CONFIG_KEY, str, default=None,
                                 description="""The location of the configuration file for your script here.
                                             Up to 100 config files will be read!""")


def load_bot_appearance(looks_config_object: ConfigObject, team_num: int) -> LoadoutConfig:
    loadout_config = LoadoutConfig()
    loadout_config.paint_config = LoadoutPaintConfig()

    loadout_header = BOT_CONFIG_LOADOUT_HEADER
    if team_num == 1 and looks_config_object.has_section(BOT_CONFIG_LOADOUT_ORANGE_HEADER):
        loadout_header = BOT_CONFIG_LOADOUT_ORANGE_HEADER

    parse_bot_loadout(loadout_config, looks_config_object, loadout_header)

    if team_num == 0 and looks_config_object.has_section(BOT_CONFIG_LOADOUT_PAINT_BLUE_HEADER):
        parse_bot_loadout_paint(loadout_config.paint_config, looks_config_object, BOT_CONFIG_LOADOUT_PAINT_BLUE_HEADER)

    if team_num == 1 and looks_config_object.has_section(BOT_CONFIG_LOADOUT_PAINT_ORANGE_HEADER):
        parse_bot_loadout_paint(loadout_config.paint_config, looks_config_object,
                                BOT_CONFIG_LOADOUT_PAINT_ORANGE_HEADER)

    return loadout_config


def create_looks_configurations() -> ConfigObject:
    config = ConfigObject()
    config.add_header(BOT_CONFIG_LOADOUT_HEADER, create_loadout())
    config.add_header(BOT_CONFIG_LOADOUT_ORANGE_HEADER, create_loadout())
    config.add_header(BOT_CONFIG_LOADOUT_PAINT_BLUE_HEADER, create_loadout_paint())
    config.add_header(BOT_CONFIG_LOADOUT_PAINT_ORANGE_HEADER, create_loadout_paint())
    return config


def create_loadout() -> ConfigHeader:
    header = ConfigHeader()
    header.add_value('team_color_id', int, default=0, description='Primary Color selection')
    header.add_value('custom_color_id', int, default=0, description='Secondary Color selection')
    header.add_value('car_id', int, default=0, description='Car type (Octane, Merc, etc)')
    header.add_value('decal_id', int, default=0, description='Type of decal')
    header.add_value('wheels_id', int, default=0, description='Wheel selection')
    header.add_value('boost_id', int, default=0, description='Boost selection')
    header.add_value('antenna_id', int, default=0, description='Antenna Selection')
    header.add_value('hat_id', int, default=0, description='Hat Selection')
    header.add_value('paint_finish_id', int, default=0, description='Paint Type (for first color)')
    header.add_value('custom_finish_id', int, default=0, description='Paint Type (for secondary color)')
    header.add_value('engine_audio_id', int, default=0, description='Engine Audio Selection')
    header.add_value('trails_id', int, default=0, description='Car trail Selection')
    header.add_value('goal_explosion_id', int, default=0, description='Goal Explosion Selection')
    header.add_value('primary_color_lookup', str, default=None,
                     description='Finds the closest primary color swatch based on the provided RGB value '
                                 'like [34, 255, 60]')
    header.add_value('secondary_color_lookup', str, default=None,
                     description='Finds the closest secondary color swatch based on the provided RGB value '
                                 'like [34, 255, 60]')

    return header


def create_loadout_paint() -> ConfigHeader:
    header = ConfigHeader()

    header.add_value('car_paint_id', int, default=0)
    header.add_value('decal_paint_id', int, default=0)
    header.add_value('wheels_paint_id', int, default=0)
    header.add_value('boost_paint_id', int, default=0)
    header.add_value('antenna_paint_id', int, default=0)
    header.add_value('hat_paint_id', int, default=0)
    header.add_value('trails_paint_id', int, default=0)
    header.add_value('goal_explosion_paint_id', int, default=0)

    return header


def parse_bot_loadout(player_configuration, bot_config, loadout_header):
    player_configuration.team_color_id = bot_config.getint(loadout_header, 'team_color_id')
    player_configuration.custom_color_id = bot_config.getint(loadout_header, 'custom_color_id')
    player_configuration.car_id = bot_config.getint(loadout_header, 'car_id')
    player_configuration.decal_id = bot_config.getint(loadout_header, 'decal_id')
    player_configuration.wheels_id = bot_config.getint(loadout_header, 'wheels_id')
    player_configuration.boost_id = bot_config.getint(loadout_header, 'boost_id')
    player_configuration.antenna_id = bot_config.getint(loadout_header, 'antenna_id')
    player_configuration.hat_id = bot_config.getint(loadout_header, 'hat_id')
    player_configuration.paint_finish_id = bot_config.getint(loadout_header, 'paint_finish_id')
    player_configuration.custom_finish_id = bot_config.getint(loadout_header, 'custom_finish_id')
    player_configuration.engine_audio_id = bot_config.getint(loadout_header, 'engine_audio_id')
    player_configuration.trails_id = bot_config.getint(loadout_header, 'trails_id')
    player_configuration.goal_explosion_id = bot_config.getint(loadout_header, 'goal_explosion_id')
    player_configuration.primary_color_lookup = parse_color_string(
        bot_config.get(loadout_header, 'primary_color_lookup'))
    player_configuration.secondary_color_lookup = parse_color_string(
        bot_config.get(loadout_header, 'secondary_color_lookup'))


def parse_color_string(color_lookup_string):
    if color_lookup_string is None:
        return None
    if color_lookup_string == 'None':
        return None
    try:
        color_array = json.loads(color_lookup_string)
        return Color(
            red=color_array[0],
            green=color_array[1],
            blue=color_array[2],
            alpha=color_array[3] if len(color_array) > 3 else 255)
    except JSONDecodeError:
        logger.warn(f"Failed to parse color lookup: {color_lookup_string}")
        return None


def parse_bot_loadout_paint(paint_config: LoadoutPaintConfig, bot_config: ConfigObject, loadout_header: str):
    paint_config.car_paint_id = bot_config.getint(loadout_header, 'car_paint_id')
    paint_config.decal_paint_id = bot_config.getint(loadout_header, 'decal_paint_id')
    paint_config.wheels_paint_id = bot_config.getint(loadout_header, 'wheels_paint_id')
    paint_config.boost_paint_id = bot_config.getint(loadout_header, 'boost_paint_id')
    paint_config.antenna_paint_id = bot_config.getint(loadout_header, 'antenna_paint_id')
    paint_config.hat_paint_id = bot_config.getint(loadout_header, 'hat_paint_id')
    paint_config.trails_paint_id = bot_config.getint(loadout_header, 'trails_paint_id')
    paint_config.goal_explosion_paint_id = bot_config.getint(loadout_header, 'goal_explosion_paint_id')



from enum import IntEnum
from socket import socket, SHUT_RDWR
from time import sleep
from typing import List, Callable, TYPE_CHECKING

from rlbot.utils.game_state_util import GameState

from flatbuffers.builder import Builder
from rlbot.agents.base_agent import SimpleControllerState
if TYPE_CHECKING:
    from rlbot.matchconfig.match_config import MatchConfig
from rlbot.messages.flat import ReadyMessage
from rlbot.messages.flat.BallPrediction import BallPrediction
from rlbot.messages.flat.DesiredGameState import DesiredGameState
from rlbot.messages.flat.FieldInfo import FieldInfo
from rlbot.messages.flat.GameMessage import GameMessage
from rlbot.messages.flat.GameTickPacket import GameTickPacket
from rlbot.messages.flat.MatchSettings import MatchSettings
from rlbot.messages.flat.MessagePacket import MessagePacket
from rlbot.messages.flat.PlayerInputChange import PlayerInputChange
from rlbot.messages.flat.PlayerSpectate import PlayerSpectate
from rlbot.messages.flat.PlayerStatEvent import PlayerStatEvent
from rlbot.messages.flat.QuickChat import QuickChat
from rlbot.utils.logging_utils import get_logger


class SocketDataType(IntEnum):
    GAME_TICK_PACKET = 1
    FIELD_INFO = 2
    MATCH_SETTINGS = 3
    PLAYER_INPUT = 4
    ACTOR_MAPPING = 5
    COMPUTER_ID = 6
    DESIRED_GAME_STATE = 7
    RENDER_GROUP = 8
    QUICK_CHAT = 9
    BALL_PREDICTION = 10
    READY_MESSAGE = 11
    MESSAGE_PACKET = 12


def int_to_bytes(val: int) -> bytes:
    return val.to_bytes(2, byteorder='big')


def int_from_bytes(bytes: bytes) -> int:
    return int.from_bytes(bytes, 'big')


class SocketMessage:
    def __init__(self, type: SocketDataType, data: bytes):
        self.type = type
        self.data = data


def read_from_socket(s: socket) -> SocketMessage:
    type_int = int_from_bytes(s.recv(2))
    data_type = SocketDataType(type_int)
    size = int_from_bytes(s.recv(2))
    data = s.recv(size)
    return SocketMessage(data_type, data)


class SocketRelay:
    def __init__(self, connection_timeout:float=120):
        self.connection_timeout = connection_timeout
        self.logger = get_logger('socket_man')
        self.socket = socket()
        self.is_connected = False
        self._should_continue = True
        self.on_connect_handlers: List[Callable[[], None]] = []
        self.packet_handlers: List[Callable[[GameTickPacket], None]] = []
        self.field_info_handlers: List[Callable[[FieldInfo], None]] = []
        self.match_settings_handlers: List[Callable[[MatchSettings], None]] = []
        self.quick_chat_handlers: List[Callable[[QuickChat], None]] = []
        self.ball_prediction_handlers: List[Callable[[BallPrediction], None]] = []
        self.player_input_change_handlers: List[Callable[[PlayerInputChange, float, int], None]] = []
        self.player_stat_handlers: List[Callable[[PlayerStatEvent, float, int], None]] = []
        self.player_spectate_handlers: List[Callable[[PlayerSpectate, float, int], None]] = []

    def send_flatbuffer(self, builder: Builder, data_type: SocketDataType):
        flatbuffer_bytes = builder.Output()
        size = len(flatbuffer_bytes)
        message = int_to_bytes(data_type) + int_to_bytes(size) + flatbuffer_bytes
        self.socket.sendall(message)

    def send_player_input(self, player_index: int, input: SimpleControllerState):
        builder = input.to_flatbuffer(player_index)
        self.send_flatbuffer(builder, SocketDataType.PLAYER_INPUT)

    def send_match_config(self, mc: 'MatchConfig'):
        builder = mc.create_flatbuffer()
        self.send_flatbuffer(builder, SocketDataType.MATCH_SETTINGS)

    def send_game_state(self, state: GameState):
        builder = Builder(400)
        game_state_offset = state.convert_to_flat(builder)
        if game_state_offset is None:
            return  # There are no values to be set, so just skip it
        builder.Finish(game_state_offset)
        self.send_flatbuffer(builder, SocketDataType.DESIRED_GAME_STATE)
        pass

    def connect_and_run(self, wants_quick_chat, wants_game_messages, wants_ball_predictions):
        """
        Connects to the socket and begins a loop that reads messages and calls any handlers
        that have been registered. Connect and run are combined into a single method because
        currently bad things happen if the buffer is allowed to fill up.
        """
        self.socket.settimeout(self.connection_timeout)
        for i in range(int(self.connection_timeout * 10)):
            try:
                self.socket.connect(('127.0.0.1', 23234))
                break
            except ConnectionRefusedError:
                sleep(.1)

        self.socket.settimeout(None)
        self.is_connected = True
        self.logger.info("Connected!")
        for handler in self.on_connect_handlers:
            handler()

        builder = self.make_ready_message(wants_ball_predictions, wants_game_messages, wants_quick_chat)
        self.send_flatbuffer(builder, SocketDataType.READY_MESSAGE)

        while self._should_continue:
            incoming_message = read_from_socket(self.socket)
            self.handle_incoming_message(incoming_message)

        # Sleep to give a chance for TCP communications to finish being acknowledged
        sleep(0.5)
        self.socket.shutdown(SHUT_RDWR)
        self.socket.close()

    def make_ready_message(self, wants_ball_predictions, wants_game_messages, wants_quick_chat):
        builder = Builder(50)
        ReadyMessage.ReadyMessageStart(builder)
        ReadyMessage.ReadyMessageAddWantsBallPredictions(builder, wants_ball_predictions)
        ReadyMessage.ReadyMessageAddWantsQuickChat(builder, wants_quick_chat)
        ReadyMessage.ReadyMessageAddWantsGameMessages(builder, wants_game_messages)
        offset = ReadyMessage.ReadyMessageEnd(builder)
        builder.Finish(offset)
        return builder

    def handle_incoming_message(self, incoming_message):
        if incoming_message.type == SocketDataType.GAME_TICK_PACKET and len(self.packet_handlers) > 0:
            packet = GameTickPacket.GetRootAsGameTickPacket(incoming_message.data, 0)
            for handler in self.packet_handlers:
                handler(packet)
        elif incoming_message.type == SocketDataType.FIELD_INFO and len(self.field_info_handlers) > 0:
            field_info = FieldInfo.GetRootAsFieldInfo(incoming_message.data, 0)
            for handler in self.field_info_handlers:
                handler(field_info)
        elif incoming_message.type == SocketDataType.MATCH_SETTINGS and len(self.match_settings_handlers) > 0:
            match_settings = MatchSettings.GetRootAsMatchSettings(incoming_message.data, 0)
            for handler in self.match_settings_handlers:
                handler(match_settings)
        elif incoming_message.type == SocketDataType.QUICK_CHAT and len(self.quick_chat_handlers) > 0:
            quick_chat = QuickChat.GetRootAsQuickChat(incoming_message.data, 0)
            for handler in self.quick_chat_handlers:
                handler(quick_chat)
        elif incoming_message.type == SocketDataType.BALL_PREDICTION and len(self.ball_prediction_handlers) > 0:
            ball_prediction = BallPrediction.GetRootAsBallPrediction(incoming_message.data, 0)
            for handler in self.ball_prediction_handlers:
                handler(ball_prediction)
        elif incoming_message.type == SocketDataType.MESSAGE_PACKET:
            if len(self.player_stat_handlers) > 0 or len(self.player_input_change_handlers) > 0 or len(
                    self.player_spectate_handlers) > 0:
                msg_packet = MessagePacket.GetRootAsMessagePacket(incoming_message.data, 0)
                for i in range(msg_packet.MessagesLength()):
                    msg = msg_packet.Messages(i)
                    msg_type = msg.MessageType()
                    if msg_type == GameMessage.PlayerSpectate and len(self.player_spectate_handlers) > 0:
                        spectate = PlayerSpectate()
                        spectate.Init(msg.Message().Bytes, msg.Message().Pos)
                        for handler in self.player_spectate_handlers:
                            handler(spectate, msg_packet.GameSeconds(), msg_packet.FrameNum())
                    elif msg_type == GameMessage.PlayerStatEvent and len(self.player_stat_handlers) > 0:
                        stat_event = PlayerStatEvent()
                        stat_event.Init(msg.Message().Bytes, msg.Message().Pos)
                        for handler in self.player_stat_handlers:
                            handler(stat_event, msg_packet.GameSeconds(), msg_packet.FrameNum())
                    elif msg_type == GameMessage.PlayerInputChange and len(
                            self.player_input_change_handlers) > 0:
                        input_change = PlayerInputChange()
                        input_change.Init(msg.Message().Bytes, msg.Message().Pos)
                        for handler in self.player_input_change_handlers:
                            handler(input_change, msg_packet.GameSeconds(), msg_packet.FrameNum())

    def disconnect(self):
        self._should_continue = False



import multiprocessing as mp
import os
import sys
import json
from dataclasses import dataclass

import psutil
import platform
import queue
import subprocess
import time
import webbrowser
from contextlib import contextmanager
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional, Dict, Set
from urllib.parse import ParseResult as URL

from dataclasses import dataclass

from rlbot.gamelaunch.epic_launch import launch_with_epic_login_trick, launch_with_epic_simple
from rlbot.socket.socket_data_reporter import get_one_packet
from rlbot.utils.structures import game_data_struct

from rlbot import gateway_util
from rlbot import version
from rlbot.base_extension import BaseExtension
from rlbot.botmanager.bot_manager_independent import BotManagerIndependent
from rlbot.botmanager.bot_manager_struct import BotManagerStruct
from rlbot.botmanager.helper_process_manager import HelperProcessManager
from rlbot.botmanager.agent_metadata import AgentMetadata
from rlbot.gateway_util import LaunchOptions, NetworkingRole
from rlbot.matchconfig.conversions import parse_match_config
from rlbot.matchconfig.match_config import MatchConfig, PlayerConfig
from rlbot.matchconfig.psyonix_config import set_random_psyonix_bot_preset
from rlbot.matchcomms.server import launch_matchcomms_server
from rlbot.parsing.bot_config_bundle import get_bot_config_bundle, BotConfigBundle, get_script_config_bundle
from rlbot.parsing.custom_config import ConfigObject
from rlbot.parsing.rlbot_config_parser import create_bot_config_layout
from rlbot.utils import process_configuration
from rlbot.utils.class_importer import import_class_with_base, import_agent
from rlbot.utils.logging_utils import get_logger, DEFAULT_LOGGER
from rlbot.utils.process_configuration import WrongProcessArgs
from rlbot.utils.structures.start_match_structures import MAX_PLAYERS
from rlbot.utils.structures.game_interface import GameInterface
from rlbot.utils.config_parser import mergeTASystemSettings, cleanUpTASystemSettings
from rlbot.matchcomms.server import MatchcommsServerThread
from rlbot.utils.virtual_environment_management import EnvBuilderWithRequirements

if platform.system() == 'Windows':
    import msvcrt

# By default, look for rlbot.cfg in the current working directory.
DEFAULT_RLBOT_CONFIG_LOCATION = os.path.realpath('./rlbot.cfg')
RLBOT_CONFIGURATION_HEADER = 'RLBot Configuration'


class ROCKET_LEAGUE_PROCESS_INFO:
    GAMEID = 252950
    PROGRAM_NAME = 'RocketLeague.exe'
    PROGRAM = 'RocketLeague.exe'
    REQUIRED_ARGS = {r'-rlbot', r'RLBot_ControllerURL=127.0.0.1:[0-9]+'}

    @staticmethod
    def get_ideal_args(port):
        # We are specifying RLBot_PacketSendRate=240, which will override people's TARLBot.ini settings.
        # We believe there is no downside to 240. See https://github.com/RLBot/RLBot/wiki/Tick-Rate
        return ['-rlbot', f'RLBot_ControllerURL=127.0.0.1:{port}', 'RLBot_PacketSendRate=240', '-nomovie']

@dataclass
class RocketLeagueLauncherPreference:
    STEAM = 'steam'
    EPIC = 'epic'
    preferred_launcher: str
    use_login_tricks: bool


# By default, we will attempt Epic with no login tricks, then fall back to Steam.
DEFAULT_LAUNCHER_PREFERENCE = RocketLeagueLauncherPreference(RocketLeagueLauncherPreference.EPIC, False)


@contextmanager
def setup_manager_context():
    """
    Creates a initialized context manager which shuts down at the end of the
    `with` block.

    usage:
    >>> with setup_manager_context() as setup_manager:
    ...     setup_manager.load_config(...)
    ...     # ... Run match
    """
    setup_manager = SetupManager()
    setup_manager.connect_to_game()
    try:
        yield setup_manager
    except Exception as e:
        get_logger(DEFAULT_LOGGER).error(e)
        raise e
    finally:
        setup_manager.shut_down(kill_all_pids=True)


@dataclass
class BotProcessInfo:
    process: mp.Process
    subprocess: subprocess.Popen
    player_config: PlayerConfig

    def is_alive(self):
        if self.process is not None:
            return self.process.is_alive()
        return self.subprocess.poll() is None


class SetupManager:
    """
    This class is responsible for pulling together all bits of the framework to
    set up a match between agents.

    A normal order of methods would be:
        connect_to_game()
        load_config()
        launch_ball_prediction()
        launch_bot_processes()
        start_match()
        infinite_loop()
        # the below two might be from another thread
        reload_all_agents()
        shut_down()
    """
    has_started = False
    num_participants = None
    names = None
    teams = None
    python_files = None
    bot_bundles: List[BotConfigBundle] = None
    match_config: MatchConfig = None
    agent_metadata_queue = mp.Queue()
    extension = None
    bot_processes: Dict[int, BotProcessInfo] = {}
    script_processes: Dict[int, subprocess.Popen] = {}

    def __init__(self):
        self.logger = get_logger(DEFAULT_LOGGER)
        self.game_interface = GameInterface(self.logger)
        self.quit_event = mp.Event()
        self.helper_process_manager = HelperProcessManager(self.quit_event)
        self.bot_quit_callbacks = []
        self.bot_reload_requests = []
        self.agent_metadata_map: Dict[int, AgentMetadata] = {}
        self.match_config: MatchConfig = None
        self.rlbot_gateway_process = None
        self.matchcomms_server: MatchcommsServerThread = None
        self.early_start_seconds = 0
        self.num_metadata_received = 0

    def is_rocket_league_running(self, port) -> bool:
        """
        Returns whether Rocket League is running with the right port.
        """

        try:
            is_rocket_league_running, proc = process_configuration.is_process_running(
                ROCKET_LEAGUE_PROCESS_INFO.PROGRAM,
                ROCKET_LEAGUE_PROCESS_INFO.PROGRAM_NAME,
                ROCKET_LEAGUE_PROCESS_INFO.REQUIRED_ARGS)

            if proc is not None:
                # Check for correct port.
                rocket_league_port = self._read_port_from_rocket_league_args(proc.cmdline())
                if rocket_league_port is not None and rocket_league_port != port:
                    raise Exception(f"Rocket League is already running with port {rocket_league_port} but we wanted "
                                    f"{port}! Please close Rocket League and let us start it for you instead!")
        except WrongProcessArgs:
            raise Exception(f"Rocket League is not running with {ROCKET_LEAGUE_PROCESS_INFO.REQUIRED_ARGS}!\n"
                            "Please close Rocket League and let us start it for you instead!")

        return is_rocket_league_running

    def connect_to_game(self, launcher_preference: RocketLeagueLauncherPreference = DEFAULT_LAUNCHER_PREFERENCE):
        """
        Connects to the game by initializing self.game_interface.
        """
        version.print_current_release_notes()
        port = self.ensure_rlbot_gateway_started()

        # Prevent loading game interface twice.
        if self.has_started:
            if not self.is_rocket_league_running(port):
                raise Exception("Rocket League is not running even though we started it once.\n"
                                "Please restart RLBot.")
            return

        # Currently match_config is None when launching from RLBotGUI.
        if self.match_config is not None and self.match_config.networking_role == 'remote_rlbot_client':
            self.logger.info("Will not start Rocket League because this is configured as a client!")
        # Launch the game if it is not running.
        elif not self.is_rocket_league_running(port):
            mergeTASystemSettings()
            self.launch_rocket_league(port=port, launcher_preference=launcher_preference)

        self.has_started = True

    @staticmethod
    def _read_port_from_rocket_league_args(args):
        for arg in args:
            # The arg will look like RLBot_ControllerURL="127.0.0.1:23233"
            if 'RLBot_ControllerURL' in arg:
                rocket_league_port = int(arg.split(':')[1].replace('"', ''))
                return int(rocket_league_port)
        return None

    def launch_rocket_league(self, port, launcher_preference: RocketLeagueLauncherPreference = DEFAULT_LAUNCHER_PREFERENCE):
        """
        Launches Rocket League but does not connect to it.
        """
        ideal_args = ROCKET_LEAGUE_PROCESS_INFO.get_ideal_args(port)

        if launcher_preference.preferred_launcher == RocketLeagueLauncherPreference.EPIC:
            if launcher_preference.use_login_tricks:
                if launch_with_epic_login_trick(ideal_args):
                    return
                else:
                    self.logger.info("Epic login trick seems to have failed, falling back to simple Epic launch.")
            # Fall back to simple if the tricks failed or we opted out of tricks.
            if launch_with_epic_simple(ideal_args):
                return

        # Try launch via Steam.
        steam_exe_path = try_get_steam_executable_path()
        if steam_exe_path:  # Note: This Python 3.8 feature would be useful here https://www.python.org/dev/peps/pep-0572/#abstract
            exe_and_args = [
                str(steam_exe_path),
                '-applaunch',
                str(ROCKET_LEAGUE_PROCESS_INFO.GAMEID),
            ] + ideal_args
            self.logger.info(f'Launching Rocket League with: {exe_and_args}')
            _ = subprocess.Popen(exe_and_args)  # This is deliberately an orphan process.
            return

        self.logger.warning(f'Launching Rocket League using Steam-only fall-back launch method with args: {ideal_args}')
        self.logger.info("You should see a confirmation pop-up, if you don't see it then click on Steam! "
                         'https://gfycat.com/AngryQuickFinnishspitz')
        args_string = '%20'.join(ideal_args)

        # Try launch via terminal (Linux)
        if platform.system() == 'Linux':
            linux_args = [
                'steam',
                f'steam://rungameid/{ROCKET_LEAGUE_PROCESS_INFO.GAMEID}//{args_string}'
            ]

            try:
                _ = subprocess.Popen(linux_args)
                return
            except OSError:
                self.logger.warning('Could not launch Steam executable on Linux.')

        try:
            self.logger.info("Launching rocket league via steam browser URL as a last resort...")
            webbrowser.open(f'steam://rungameid/{ROCKET_LEAGUE_PROCESS_INFO.GAMEID}//{args_string}')
        except webbrowser.Error:
            self.logger.warning(
                'Unable to launch Rocket League. Please launch Rocket League manually using the -rlbot option to continue.')

    def load_match_config(self, match_config: MatchConfig, bot_config_overrides={}):
        """
        Loads the match config into internal data structures, which prepares us to later
        launch bot processes and start the match.

        This is an alternative to the load_config method; they accomplish the same thing.
        """
        self.num_participants = match_config.num_players
        self.names = [bot.name for bot in match_config.player_configs]
        self.teams = [bot.team for bot in match_config.player_configs]

        for player in match_config.player_configs:
            if player.bot and not player.rlbot_controlled and not player.loadout_config:
                set_random_psyonix_bot_preset(player)

        bundles = [bot_config_overrides[index] if index in bot_config_overrides else
                   get_bot_config_bundle(bot.config_path) if bot.config_path else None
                   for index, bot in enumerate(match_config.player_configs)]

        self.python_files = [bundle.python_file if bundle else None
                             for bundle in bundles]

        self.bot_bundles = []

        for index, bot in enumerate(match_config.player_configs):
            self.bot_bundles.append(bundles[index])
            if bot.loadout_config is None and bundles[index]:
                bot.loadout_config = bundles[index].generate_loadout_config(index, bot.team)

        if match_config.extension_config is not None and match_config.extension_config.python_file_path is not None:
            self.load_extension(match_config.extension_config.python_file_path)

        for bundle in self.bot_bundles:
            if bundle is not None and bundle.use_virtual_environment:
                builder = EnvBuilderWithRequirements(bundle=bundle)
                builder.create(Path(bundle.config_directory) / 'venv')

        for script_config in match_config.script_configs:
            script_config_bundle = get_script_config_bundle(script_config.config_path)
            if script_config_bundle.use_virtual_environment:
                builder = EnvBuilderWithRequirements(bundle=script_config_bundle)
                builder.create(Path(script_config_bundle.config_directory) / 'venv')

        self.match_config = match_config
        self.game_interface.match_config = match_config
        self.game_interface.start_match_flatbuffer = match_config.create_flatbuffer()

    def load_config(self, framework_config: ConfigObject = None, config_location=DEFAULT_RLBOT_CONFIG_LOCATION,
                    bot_configs=None,
                    looks_configs=None):
        """
        Loads the configuration into internal data structures, which prepares us to later
        launch bot processes and start the match.

        :param framework_config: A config object that indicates what bots to run. May come from parsing a rlbot.cfg.
        :param config_location: The location of the rlbot.cfg file, which will be used to resolve relative paths.
        :param bot_configs: Overrides for bot configurations.
        :param looks_configs: Overrides for looks configurations.
        """
        self.logger.debug('reading the configs')

        # Set up RLBot.cfg
        if framework_config is None:
            framework_config = create_bot_config_layout()
            framework_config.parse_file(config_location, max_index=MAX_PLAYERS)
        if bot_configs is None:
            bot_configs = {}
        if looks_configs is None:
            looks_configs = {}

        match_config = parse_match_config(framework_config, config_location, bot_configs, looks_configs)
        self.load_match_config(match_config, bot_configs)

    def ensure_rlbot_gateway_started(self) -> int:
        """
        Ensures that RLBot.exe is running. Returns the port that it will be listening on for connections from
        Rocket League. Rocket League should be passed a command line argument so that it starts with this same port.
        :return:
        """

        # TODO: Uncomment this when done with local testing of Remote RLBot.
        self.rlbot_gateway_process, port = gateway_util.find_existing_process()
        if self.rlbot_gateway_process is not None:
            self.logger.info(f"Already have RLBot.exe running! Port is {port}")
            return port

        launch_options = LaunchOptions()
        if self.match_config is not None:  # Currently this is None when launching from RLBotGUI.
            networking_role = NetworkingRole[self.match_config.networking_role]
            launch_options = LaunchOptions(
                networking_role=networking_role,
                remote_address=self.match_config.network_address)

        self.rlbot_gateway_process, port = gateway_util.launch(launch_options)
        self.logger.info(f"Python started RLBot.exe with process id {self.rlbot_gateway_process.pid} "
                         f"and port {port}")
        return port

    def launch_ball_prediction(self):
        # This does nothing now. It's kept here temporarily so that RLBotGUI doesn't break.
        pass

    def has_received_metadata_from_all_bots(self):
        expected_metadata_calls = sum(1 for player in self.match_config.player_configs if player.rlbot_controlled)
        return self.num_metadata_received >= expected_metadata_calls

    def launch_early_start_bot_processes(self, match_config: MatchConfig = None):
        """
        Some bots can start up before the game is ready and not be bothered by missing
        or strange looking values in the game tick packet, etc. Such bots can opt in to the
        early start category and enjoy extra time to load up before the match starts.

        WARNING: Early start is a bad idea if there's any risk that bots will not get their promised
        index. This can happen with remote RLBot, etc.
        """

        if self.match_config.networking_role == NetworkingRole.remote_rlbot_client:
            return  # The bot indices are liable to change, so don't start anything yet.

        self.logger.debug("Launching early-start bot processes")
        num_started = self.launch_bot_process_helper(early_starters_only=True, match_config=match_config or self.match_config)
        self.try_recieve_agent_metadata()
        if num_started > 0 and self.early_start_seconds > 0:
            self.logger.info(f"Waiting for {self.early_start_seconds} seconds to let early-start bots load.")
            end_time = datetime.now() + timedelta(seconds=self.early_start_seconds)
            while datetime.now() < end_time:
                self.try_recieve_agent_metadata()
                time.sleep(0.1)

    def launch_bot_processes(self, match_config: MatchConfig = None):
        self.logger.debug("Launching bot processes")
        self.launch_bot_process_helper(early_starters_only=False, match_config=match_config or self.match_config)

    def launch_bot_process_helper(self, early_starters_only=False, match_config: MatchConfig = None):
        # Start matchcomms here as it's only required for the bots.
        self.kill_matchcomms_server()
        self.matchcomms_server = launch_matchcomms_server()
        self.bot_processes = {ind: proc for ind, proc in self.bot_processes.items() if proc.is_alive()}

        num_started = 0

        # Launch processes
        # TODO: this might be the right moment to fix the player indices based on a game tick packet.
        if not early_starters_only:
            packet = get_one_packet()

        # TODO: root through the packet and find discrepancies in the player index mapping.
        for i in range(min(self.num_participants, len(match_config.player_configs))):

            player_config = match_config.player_configs[i]
            if not player_config.has_bot_script():
                continue
            if early_starters_only and not self.bot_bundles[i].supports_early_start:
                continue

            spawn_id = player_config.spawn_id

            if early_starters_only:
                # Danger: we have low confidence in this since we're not leveraging the spawn id.
                participant_index = i
            else:
                participant_index = None

                self.logger.info(f'Player in slot {i} was sent with spawn id {spawn_id}, will search in the packet.')
                for n in range(0, packet.PlayersLength()):
                    packet_spawn_id = packet.Players(n).SpawnId()
                    if spawn_id == packet_spawn_id:
                        self.logger.info(f'Looks good, considering participant index to be {n}')
                        participant_index = n
                if participant_index is None:
                    for prox_index, proc_info in self.bot_processes.items():
                        if spawn_id == proc_info.player_config.spawn_id:
                            participant_index = prox_index
                    if participant_index is None:
                        raise Exception(f"Unable to determine the bot index for spawn id {spawn_id}")

            if participant_index not in self.bot_processes:
                bundle = get_bot_config_bundle(player_config.config_path)
                name = str(self.match_config.player_configs[i].name)
                if bundle.supports_standalone:
                    executable = sys.executable
                    if bundle.use_virtual_environment:
                        executable = str(Path(bundle.config_directory) / 'venv' / 'Scripts' / 'python.exe')
                    process = subprocess.Popen([
                        executable,
                        bundle.python_file,
                        '--config-file', str(player_config.config_path),
                        '--name', name,
                        '--team', str(self.teams[i]),
                        '--player-index', str(participant_index),
                        '--spawn-id', str(spawn_id),
                        '--matchcomms-url', self.matchcomms_server.root_url.geturl()
                    ], cwd=Path(bundle.config_directory).parent)
                    self.bot_processes[participant_index] = BotProcessInfo(process=None, subprocess=process, player_config=player_config)

                    # Insert immediately into the agent metadata map because the standalone process has no way to communicate it back out
                    self.agent_metadata_map[participant_index] = AgentMetadata(participant_index, name, self.teams[i], {process.pid})
                else:
                    reload_request = mp.Event()
                    quit_callback = mp.Event()
                    self.bot_reload_requests.append(reload_request)
                    self.bot_quit_callbacks.append(quit_callback)
                    process = mp.Process(target=SetupManager.run_agent,
                                         args=(self.quit_event, quit_callback, reload_request, self.bot_bundles[i],
                                               name,
                                               self.teams[i], participant_index, self.python_files[i], self.agent_metadata_queue,
                                               match_config, self.matchcomms_server.root_url, spawn_id))
                    process.start()
                    self.bot_processes[participant_index] = BotProcessInfo(process=process, subprocess=None, player_config=player_config)
                num_started += 1

        self.logger.info(f"Successfully started {num_started} bot processes")

        process_configuration.configure_processes(self.agent_metadata_map, self.logger)

        scripts_started = 0
        for script_config in match_config.script_configs:
            script_config_bundle = get_script_config_bundle(script_config.config_path)
            if early_starters_only and not script_config_bundle.supports_early_start:
                continue
            executable = sys.executable
            if script_config_bundle.use_virtual_environment:
                executable = str(Path(script_config_bundle.config_directory) / 'venv' / 'Scripts' / 'python.exe')

            process = subprocess.Popen([executable, script_config_bundle.script_file],
                                       cwd=Path(script_config_bundle.config_directory).parent)
            self.logger.info(f"Started script with pid {process.pid} using {process.args}")
            self.script_processes[process.pid] = process
            scripts_started += 1

        self.logger.debug(f"Successfully started {scripts_started} scripts")

        return num_started

    def launch_quick_chat_manager(self):
        # Quick chat manager is gone since we're using RLBot.exe now.
        # Keeping this function around for backwards compatibility.
        pass

    def start_match(self):

        if self.match_config.networking_role == NetworkingRole.remote_rlbot_client:
            match_settings = self.game_interface.get_match_settings()
            # TODO: merge the match settings into self.match_config
            # And then make sure we still only start the appropriate bot processes
            # that we originally asked for.

        self.logger.info("Python attempting to start match.")
        self.game_interface.start_match()
        self.game_interface.wait_until_valid_packet()
        self.logger.info("Match has started")

        cleanUpTASystemSettings()

    def infinite_loop(self):
        instructions = "Press 'r' to reload all agents, or 'q' to exit"
        self.logger.info(instructions)
        while not self.quit_event.is_set():
            # Handle commands
            # TODO windows only library
            if platform.system() == 'Windows':
                if msvcrt.kbhit():
                    command = msvcrt.getwch()
                    if command.lower() == 'r':  # r: reload
                        self.reload_all_agents()
                    elif command.lower() == 'q' or command == '\u001b':  # q or ESC: quit
                        self.shut_down()
                        break
                    # Print instructions again if a alphabet character was pressed but no command was found
                    elif command.isalpha():
                        self.logger.info(instructions)
            else:
                try:
                    # https://python-forum.io/Thread-msvcrt-getkey-for-linux
                    import termios, sys
                    TERMIOS = termios

                    fd = sys.stdin.fileno()
                    old = termios.tcgetattr(fd)
                    new = termios.tcgetattr(fd)
                    new[3] = new[3] & ~TERMIOS.ICANON & ~TERMIOS.ECHO
                    new[6][TERMIOS.VMIN] = 1
                    new[6][TERMIOS.VTIME] = 0
                    termios.tcsetattr(fd, TERMIOS.TCSANOW, new)
                    command = None
                    try:
                        command = os.read(fd, 1)
                    finally:
                        termios.tcsetattr(fd, TERMIOS.TCSAFLUSH, old)
                    command = command.decode("utf-8")
                    if command.lower() == 'r':  # r: reload
                        self.reload_all_agents()
                    elif command.lower() == 'q' or command == '\u001b':  # q or ESC: quit
                        self.shut_down()
                        break
                    # Print instructions again if a alphabet character was pressed but no command was found
                    elif command.isalpha():
                        self.logger.info(instructions)
                except:
                    pass

            self.try_recieve_agent_metadata()

    def try_recieve_agent_metadata(self):
        """
        Checks whether any of the started bots have posted their AgentMetadata
        yet. If so, we put them on the agent_metadata_map such that we can
        kill their process later when we shut_down(kill_agent_process_ids=True)

        Returns how from how many bots we received metadata from.
        """
        num_recieved = 0
        while True:  # will exit on queue.Empty
            try:
                single_agent_metadata: AgentMetadata = self.agent_metadata_queue.get(timeout=0.1)
                num_recieved += 1
                self.helper_process_manager.start_or_update_helper_process(single_agent_metadata)
                self.agent_metadata_map[single_agent_metadata.index] = single_agent_metadata
                process_configuration.configure_processes(self.agent_metadata_map, self.logger)
            except queue.Empty:
                self.num_metadata_received += num_recieved
                return num_recieved

    def reload_all_agents(self, quiet=False):
        if not quiet:
            self.logger.info("Reloading all agents...")
        for rr in self.bot_reload_requests:
            rr.set()

    def shut_down(self, time_limit=5, kill_all_pids=False, quiet=False):
        if not quiet:
            self.logger.info("Shutting Down")

        self.quit_event.set()
        end_time = datetime.now() + timedelta(seconds=time_limit)

        # Don't kill RLBot.exe. It needs to keep running because if we're in a GUI
        # that will persist after this shut down, the interface dll in charge of starting
        # matches is already locked in to its shared memory files, and if we start a new
        # RLBot.exe, those files will go stale. https://github.com/skyborgff/RLBot/issues/9

        # Wait for all processes to terminate before terminating main process
        terminated = False
        while not terminated:
            terminated = True
            for callback in self.bot_quit_callbacks:
                if not callback.is_set():
                    terminated = False
            time.sleep(0.1)
            if datetime.now() > end_time:
                self.logger.info("Taking too long to quit, trying harder...")
                break

        self.kill_bot_processes()
        self.kill_agent_process_ids(set(self.script_processes.keys()))

        if kill_all_pids:
            # The original meaning of the kill_all_pids flag only applied to bots, not scripts,
            # so we are doing that separately.
            self.kill_agent_process_ids(process_configuration.extract_all_pids(self.agent_metadata_map))

        self.kill_matchcomms_server()

        # The quit event can only be set once. Let's reset to our initial state
        self.quit_event = mp.Event()
        self.helper_process_manager = HelperProcessManager(self.quit_event)

        if not quiet:
            self.logger.info("Shut down complete!")

    def load_extension(self, extension_filename):
        try:
            extension_class = import_class_with_base(extension_filename, BaseExtension).get_loaded_class()
            self.extension = extension_class(self)
            self.game_interface.set_extension(self.extension)
        except FileNotFoundError as e:
            print(f'Failed to load extension: {e}')

    @staticmethod
    def run_agent(terminate_event, callback_event, reload_request, bundle: BotConfigBundle, name, team, index,
                  python_file, agent_telemetry_queue, match_config: MatchConfig, matchcomms_root: URL, spawn_id: str):

        # Set the working directory to one level above the bot cfg file.
        # This mimics the behavior you get when executing run.py in one of the
        # example bot repositories, so bots will be more likely to 'just work'
        # even if the developer is careless about file paths.
        os.chdir(Path(bundle.config_directory).parent)

        agent_class_wrapper = import_agent(python_file)
        config_file = agent_class_wrapper.get_loaded_class().base_create_agent_configurations()
        config_file.parse_file(bundle.config_obj, config_directory=bundle.config_directory)

        if hasattr(agent_class_wrapper.get_loaded_class(), "run_independently"):
            bm = BotManagerIndependent(terminate_event, callback_event, reload_request, config_file, name, team, index,
                                       agent_class_wrapper, agent_telemetry_queue, match_config, matchcomms_root,
                                       spawn_id)
        else:
            bm = BotManagerStruct(terminate_event, callback_event, reload_request, config_file, name, team, index,
                                  agent_class_wrapper, agent_telemetry_queue, match_config, matchcomms_root, spawn_id)
        bm.run()

    def kill_bot_processes(self):
        for process_info in self.bot_processes.values():
            proc = process_info.process or process_info.subprocess
            proc.terminate()
        for process_info in self.bot_processes.values():
            if process_info.process:
                process_info.process.join(timeout=1)
        self.bot_processes.clear()
        self.num_metadata_received = 0

    def kill_agent_process_ids(self, pids: Set[int]):
        for pid in pids:
            try:
                parent = psutil.Process(pid)
                for child in parent.children(recursive=True):  # or parent.children() for recursive=False
                    self.logger.info(f"Killing {child.pid} (child of {pid})")
                    try:
                        child.kill()
                    except psutil.NoSuchProcess:
                        self.logger.info("Already dead.")
                self.logger.info(f"Killing {pid}")
                try:
                    parent.kill()
                except psutil.NoSuchProcess:
                    self.logger.info("Already dead.")
            except psutil.NoSuchProcess:
                self.logger.info("Can't fetch parent process, already dead.")
            except psutil.AccessDenied as ex:
                self.logger.error(f"Access denied when trying to kill a bot pid! {ex}")
            except Exception as ex:
                self.logger.error(f"Unexpected exception when trying to kill a bot pid! {ex}")

    def kill_matchcomms_server(self):
        if self.matchcomms_server:
            self.matchcomms_server.close()
            self.matchcomms_server = None


def try_get_steam_executable_path() -> Optional[Path]:
    """
    Tries to find the path of the Steam executable.
    Has platform specific code.
    """

    try:
        from winreg import OpenKey, HKEY_CURRENT_USER, ConnectRegistry, QueryValueEx, REG_SZ
    except ImportError as e:
        return  # TODO: Linux support.

    try:
        key = OpenKey(ConnectRegistry(None, HKEY_CURRENT_USER), r'Software\Valve\Steam')
        val, val_type = QueryValueEx(key, 'SteamExe')
    except FileNotFoundError:
        return
    if val_type != REG_SZ:
        return
    return Path(val)





# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Sent when connecting to RLBot to indicate what type of messages are desired.
# /// This could be sent by a bot, or a bot manager governing several bots, an
# /// overlay, or any other utility that connects to the RLBot process.
class ReadyMessage(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsReadyMessage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ReadyMessage()
        x.Init(buf, n + offset)
        return x

    # ReadyMessage
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ReadyMessage
    def WantsBallPredictions(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # ReadyMessage
    def WantsQuickChat(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # ReadyMessage
    def WantsGameMessages(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

def ReadyMessageStart(builder): builder.StartObject(3)
def ReadyMessageAddWantsBallPredictions(builder, wantsBallPredictions): builder.PrependBoolSlot(0, wantsBallPredictions, 0)
def ReadyMessageAddWantsQuickChat(builder, wantsQuickChat): builder.PrependBoolSlot(1, wantsQuickChat, 0)
def ReadyMessageAddWantsGameMessages(builder, wantsGameMessages): builder.PrependBoolSlot(2, wantsGameMessages, 0)
def ReadyMessageEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class ScoreInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsScoreInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ScoreInfo()
        x.Init(buf, n + offset)
        return x

    # ScoreInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ScoreInfo
    def Score(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScoreInfo
    def Goals(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScoreInfo
    def OwnGoals(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScoreInfo
    def Assists(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScoreInfo
    def Saves(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScoreInfo
    def Shots(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScoreInfo
    def Demolitions(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def ScoreInfoStart(builder): builder.StartObject(7)
def ScoreInfoAddScore(builder, score): builder.PrependInt32Slot(0, score, 0)
def ScoreInfoAddGoals(builder, goals): builder.PrependInt32Slot(1, goals, 0)
def ScoreInfoAddOwnGoals(builder, ownGoals): builder.PrependInt32Slot(2, ownGoals, 0)
def ScoreInfoAddAssists(builder, assists): builder.PrependInt32Slot(3, assists, 0)
def ScoreInfoAddSaves(builder, saves): builder.PrependInt32Slot(4, saves, 0)
def ScoreInfoAddShots(builder, shots): builder.PrependInt32Slot(5, shots, 0)
def ScoreInfoAddDemolitions(builder, demolitions): builder.PrependInt32Slot(6, demolitions, 0)
def ScoreInfoEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class BoostPad(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsBoostPad(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BoostPad()
        x.Init(buf, n + offset)
        return x

    # BoostPad
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # BoostPad
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # BoostPad
    def IsFullBoost(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

def BoostPadStart(builder): builder.StartObject(2)
def BoostPadAddLocation(builder, location): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def BoostPadAddIsFullBoost(builder, isFullBoost): builder.PrependBoolSlot(1, isFullBoost, 0)
def BoostPadEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Rigid body state for a player / car in the game. Includes the latest
# /// controller input, which is otherwise difficult to correlate with consequences.
class PlayerRigidBodyState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayerRigidBodyState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PlayerRigidBodyState()
        x.Init(buf, n + offset)
        return x

    # PlayerRigidBodyState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PlayerRigidBodyState
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .RigidBodyState import RigidBodyState
            obj = RigidBodyState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # PlayerRigidBodyState
    def Input(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .ControllerState import ControllerState
            obj = ControllerState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def PlayerRigidBodyStateStart(builder): builder.StartObject(2)
def PlayerRigidBodyStateAddState(builder, state): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(state), 0)
def PlayerRigidBodyStateAddInput(builder, input): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(input), 0)
def PlayerRigidBodyStateEnd(builder): return builder.EndObject()



import glob
import os
import time
import traceback
from datetime import datetime, timedelta
from urllib.parse import ParseResult as URL

from rlbot.agents.base_agent import BaseAgent, BOT_CONFIG_MODULE_HEADER, MAXIMUM_TICK_RATE_PREFERENCE_KEY
from rlbot.botmanager.agent_metadata import AgentMetadata
from rlbot.matchconfig.match_config import MatchConfig
from rlbot.messages.flat import MatchSettings
from rlbot.utils.game_state_util import GameState
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.rendering.rendering_manager import RenderingManager
from rlbot.utils.rlbot_exception import EmptyDllResponse
from rlbot.utils.structures.bot_input_struct import PlayerInput
from rlbot.utils.structures.game_data_struct import FieldInfoPacket
from rlbot.utils.structures.game_interface import GameInterface
from rlbot.utils.structures.game_status import RLBotCoreStatus
from rlbot.utils.structures.quick_chats import send_quick_chat_flat

MAX_AGENT_CALL_PERIOD = timedelta(seconds=1.0 / 2)  # Minimum call rate when paused.
REFRESH_IN_PROGRESS = 1
REFRESH_NOT_IN_PROGRESS = 0
MAX_CARS = 10

# Backward compatibility with Python 3.6:
try:
    time.perf_counter_ns()
except AttributeError:
    time.perf_counter_ns = lambda: time.perf_counter() * 1e9


# Backward compatibility with Python 3.6:
try:
    time.perf_counter_ns()
except AttributeError:
    time.perf_counter_ns = lambda: time.perf_counter() * 1e9


class BotManager:

    def __init__(self, terminate_request_event, termination_complete_event, reload_request_event, bot_configuration,
                 name, team, index, agent_class_wrapper, agent_metadata_queue, match_config: MatchConfig,
                 matchcomms_root: URL, spawn_id: int):
        """
        :param terminate_request_event: an Event (multiprocessing) which will be set from the outside when the program is trying to terminate
        :param termination_complete_event: an Event (multiprocessing) which should be set from inside this class when termination has completed successfully
        :param reload_request_event: an Event (multiprocessing) which will be set from the outside to force a reload of the agent
        :param reload_complete_event: an Event (multiprocessing) which should be set from inside this class when reloading has completed successfully
        :param bot_configuration: parameters which will be passed to the bot's constructor
        :param name: name which will be passed to the bot's constructor. Will probably be displayed in-game.
        :param team: 0 for blue team or 1 for orange team. Will be passed to the bot's constructor.
        :param index: The player index, i.e. "this is player number <index>". Will be passed to the bot's constructor.
            Can be used to pull the correct data corresponding to the bot's car out of the game tick packet.
        :param agent_class_wrapper: The ExternalClassWrapper object that can be used to load and reload the bot
        :param agent_metadata_queue: a Queue (multiprocessing) which expects to receive AgentMetadata once available.
        :param match_config: Describes the match that is being played.
        :param matchcomms_root: The server to connect to if you want to communicate to other participants in the match.
        :param spawn_id: The identifier we expect to see in the game tick packet at our player index. If it does not
            match, then we will force the agent to retire. Pass None to opt out of this behavior.
        """
        self.terminate_request_event = terminate_request_event
        self.termination_complete_event = termination_complete_event
        self.reload_request_event = reload_request_event
        self.bot_configuration = bot_configuration
        self.name = name
        self.team = team
        self.index = index
        self.agent_class_wrapper = agent_class_wrapper
        self.agent_metadata_queue = agent_metadata_queue
        self.logger = get_logger('bot' + str(self.index))
        self.game_interface = GameInterface(self.logger)
        self.last_chat_time = time.time()
        self.chat_counter = 0
        self.reset_chat_time = True
        self.game_tick_packet = None
        self.bot_input = None
        self.ball_prediction = None
        self.rigid_body_tick = None
        self.match_config = match_config
        self.matchcomms_root = matchcomms_root
        self.last_message_index = 0
        self.agent = None
        self.agent_class_file = None
        self.last_module_modification_time = 0
        self.scan_last = 0
        self.scan_temp = 0
        self.file_iterator = None
        self.maximum_tick_rate_preference = bot_configuration.get(BOT_CONFIG_MODULE_HEADER,
                                                                  MAXIMUM_TICK_RATE_PREFERENCE_KEY)
        self.spawn_id = spawn_id
        self.spawn_id_seen = False
        self.counter = 0

    def send_quick_chat_from_agent(self, team_only, quick_chat):
        """
        Passes the agents quick chats to the game, and also to other python bots.
        This does perform limiting.
        You are limited to 5 quick chats in a 2 second period starting from the first chat.
        This means you can spread your chats out to be even within that 2 second period.
        You could spam them in the first little bit but then will be throttled.
        """

        # Send the quick chat to the game
        rlbot_status = send_quick_chat_flat(self.game_interface, self.index, self.team, team_only, quick_chat)

        if rlbot_status == RLBotCoreStatus.QuickChatRateExceeded:
            self.logger.debug('quick chat disabled')

    def load_agent(self):
        """
        Loads and initializes an agent using instance variables, registers for quick chat and sets render functions.
        :return: An instance of an agent, and the agent class file.
        """
        agent_class = self.agent_class_wrapper.get_loaded_class()
        self.agent = agent_class(self.name, self.team, self.index)
        self.agent._set_spawn_id(self.spawn_id)
        self.agent.init_match_config(self.match_config)

        self.agent.load_config(self.bot_configuration.get_header("Bot Parameters"))

        self.update_metadata_queue()
        self.set_render_manager()

        self.agent_class_file = self.agent_class_wrapper.python_file
        self.agent._register_quick_chat(self.send_quick_chat_from_agent)
        self.agent._register_field_info(self.get_field_info)
        self.agent._register_set_game_state(self.set_game_state)
        self.agent._register_ball_prediction(self.get_ball_prediction)
        self.agent._register_ball_prediction_struct(self.get_ball_prediction_struct)
        self.agent._register_get_rigid_body_tick(self.get_rigid_body_tick)
        self.agent._register_match_settings_func(self.get_match_settings)
        self.agent.matchcomms_root = self.matchcomms_root

        # Once all engine setup is done, do the agent-specific initialization, if any:
        self.agent.initialize_agent()
        return self.agent, self.agent_class_file

    def set_render_manager(self):
        """
        Sets the render manager for the agent.
        :param agent: An instance of an agent.
        """
        rendering_manager = self.game_interface.renderer.get_rendering_manager(self.index, self.team)
        self.agent._set_renderer(rendering_manager)

    def update_metadata_queue(self):
        """
        Adds a new instance of AgentMetadata into the `agent_metadata_queue` using `agent` data.
        :param agent: An instance of an agent.
        """
        pids = {os.getpid(), *self.agent.get_extra_pids()}

        helper_process_request = self.agent.get_helper_process_request()

        self.agent_metadata_queue.put(AgentMetadata(self.index, self.name, self.team, pids, helper_process_request))

    def reload_agent(self):
        """
        Reloads the agent. Can throw exceptions. External classes should use reload_event.set() instead.
        """
        self.logger.info('Reloading Agent: ' + self.agent.name)
        self.agent_class_wrapper.reload()
        old_agent = self.agent
        self.load_agent()
        self.retire_agent(old_agent)  # We do this after load_agent as load_agent might fail.

    def wait_for_full_data(self):
        for i in range(10):
            match_settings = self.get_match_settings()
            if match_settings is not None and self.get_field_info().num_goals > 0:
                return
            time.sleep(0.1)
        self.logger.error("WARNING: failed to get full match data before starting bot!")

    def run(self):
        """
        Loads interface for RLBot, prepares environment and agent, and calls the update for the agent.
        """
        self.logger.debug('initializing agent')
        self.game_interface.load_interface()
        self.wait_for_full_data()

        self.prepare_for_run()

        if self.match_config is None:
            match_settings = self.game_interface.get_match_settings()
            self.match_config = MatchConfig.from_match_settings_flatbuffer(match_settings)

        last_tick_game_time = 0  # What the tick time of the last observed tick was
        last_call_real_time = datetime.now()  # When we last called the Agent
        frame_urgency = 0  # If the bot is getting called more than its preferred max rate, urgency will go negative.

        # Get bot module
        self.load_agent()

        self.last_module_modification_time = self.check_modification_time(os.path.dirname(self.agent_class_file))

        # Run until main process tells to stop, or we detect Ctrl+C
        try:
            while not self.terminate_request_event.is_set():
                self.pull_data_from_game()

                # Run the Agent only if the game_info has updated.
                tick_game_time = self.get_game_time()
                now = datetime.now()
                should_call_while_paused = now - last_call_real_time >= MAX_AGENT_CALL_PERIOD or self.match_config.enable_lockstep

                if frame_urgency < 1 / self.maximum_tick_rate_preference:
                    # Urgency increases every frame, but don't let it build up a large backlog
                    frame_urgency += tick_game_time - last_tick_game_time

                if tick_game_time != last_tick_game_time and frame_urgency >= 0 or should_call_while_paused:
                    last_call_real_time = now
                    # Urgency decreases when a tick is processed.
                    if frame_urgency > 0:
                        frame_urgency -= 1 / self.maximum_tick_rate_preference

                    self.perform_tick()
                    self.counter += 1

                last_tick_game_time = tick_game_time
                if self.spawn_id is not None:
                    packet_spawn_id = self.get_spawn_id()
                    if self.spawn_id_seen:
                        if packet_spawn_id != self.spawn_id:
                            self.logger.warn(f"The bot's spawn id {self.spawn_id} does not match the one in the packet "
                                             f"{packet_spawn_id}, retiring!")
                            break
                    elif packet_spawn_id == self.spawn_id and self.game_tick_packet.game_info.is_round_active:
                        self.spawn_id_seen = True

        except KeyboardInterrupt:
            self.terminate_request_event.set()

        self.retire_agent(self.agent)

        # If terminated, send callback
        self.termination_complete_event.set()

    def perform_tick(self):
        # Reload the Agent if it has been modified or if reload is requested from outside.
        # But only do that every 20th tick.
        if self.agent.is_hot_reload_enabled() and self.counter % 20 == 1:
            self.hot_reload_if_necessary()
        try:
            chat_messages = self.game_interface.receive_chat(self.index, self.team, self.last_message_index)
            for i in range(0, chat_messages.MessagesLength()):
                message = chat_messages.Messages(i)
                if len(self.match_config.player_configs) > message.PlayerIndex():
                    self.agent.handle_quick_chat(
                        index=message.PlayerIndex(),
                        team=self.match_config.player_configs[message.PlayerIndex()].team,
                        quick_chat=message.QuickChatSelection())
                else:
                    self.logger.debug(f"Skipping quick chat delivery for {message.MessageIndex()} because "
                                      "we don't recognize the player index. Probably stale.")
                self.last_message_index = message.MessageIndex()
        except EmptyDllResponse:
            self.logger.debug("Empty response when reading chat!")
        # Call agent
        try:
            self.call_agent(self.agent, self.agent_class_wrapper.get_loaded_class())
        except Exception as e:
            self.logger.error("Call to agent failed:\n" + traceback.format_exc())

    def hot_reload_if_necessary(self):
        try:
            new_module_modification_time = self.check_modification_time(os.path.dirname(self.agent_class_file))
            if new_module_modification_time != self.last_module_modification_time or self.reload_request_event.is_set():
                self.reload_request_event.clear()
                self.last_module_modification_time = new_module_modification_time
                # Clear the render queue on reload.
                if hasattr(self.agent, 'renderer') and isinstance(self.agent.renderer, RenderingManager):
                    self.agent.renderer.clear_all_touched_render_groups()
                self.reload_agent()
        except FileNotFoundError:
            self.logger.error(f"Agent file {self.agent_class_file} was not found. Will try again.")
            time.sleep(0.5)
        except Exception:
            self.logger.error("Reloading the agent failed:\n" + traceback.format_exc())
            time.sleep(5)  # Avoid burning CPU, and give the user a moment to read the log

    def retire_agent(self, agent):
        # Shut down the bot by calling cleanup functions.
        if hasattr(agent, 'retire'):
            try:
                agent.retire()
            except Exception as e:
                self.logger.error("Retiring the agent failed:\n" + traceback.format_exc())
        if hasattr(agent, 'renderer') and isinstance(agent.renderer, RenderingManager):
            agent.renderer.clear_all_touched_render_groups()
        # Zero out the inputs, so it's more obvious that the bot has stopped.
        self.game_interface.update_player_input(PlayerInput(), self.index)

        # Don't trust the agent to shut down its own client in retire().
        if agent._matchcomms is not None:
            agent._matchcomms.close()

    def check_modification_time(self, directory, timeout_ms=1):
        if self.scan_last > 0 and timeout_ms is not None:
            stop_time = time.perf_counter_ns() + timeout_ms * 10**6
        else:
            stop_time = None
        if self.file_iterator is None:
            self.file_iterator = glob.iglob(f"{directory}/**/*.py", recursive=True)
        for f in self.file_iterator:
            self.scan_temp = max(self.scan_temp, os.stat(f).st_mtime)
            if stop_time is not None and time.perf_counter_ns() > stop_time:
                # Timeout exceeded. The scan will pick up from here on the next call.
                break
        else:
            # Scan finished. Update the modification time and restart the scan:
            self.scan_last, self.scan_temp = self.scan_temp, 0
            self.file_iterator = None
        return self.scan_last

    def get_field_info(self) -> FieldInfoPacket:
        field_info = FieldInfoPacket()
        self.game_interface.update_field_info_packet(field_info)
        return field_info

    def get_rigid_body_tick(self):
        """Get the most recent state of the physics engine."""
        return self.game_interface.update_rigid_body_tick(self.rigid_body_tick)

    def set_game_state(self, game_state: GameState) -> None:
        self.game_interface.set_game_state(game_state)

    def get_ball_prediction(self):
        return self.game_interface.get_ball_prediction()

    def get_match_settings(self) -> MatchSettings:
        return self.game_interface.get_match_settings()

    def get_ball_prediction_struct(self):
        raise NotImplementedError

    def prepare_for_run(self):
        raise NotImplementedError

    def call_agent(self, agent: BaseAgent, agent_class):
        raise NotImplementedError

    def get_game_time(self):
        raise NotImplementedError

    def pull_data_from_game(self):
        raise NotImplementedError

    def get_spawn_id(self):
        raise NotImplementedError



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class SeriesLengthOption(object):
    Unlimited = 0
    Three_Games = 1
    Five_Games = 2
    Seven_Games = 3




from dataclasses import dataclass
from enum import IntEnum
from typing import List


# This file was created by hand for the sole purpose of giving bot makers better autocomplete.
# The ctypes structures in game_data_struct.py are what actually transport the data, but they
# fail to give autocomplete in some code editors.

# In some places we will use these classes in our type hints, but ONLY in type hints. These classes
# should probably never be constructed.

@dataclass
class Vector3:
    """
    Note that in Rocket League, the x-axis is backwards from what you might expect.
    The orange goal is in the positive y direction. To the left of the orange goal
    is the positive x direction (yes really). Be careful with trig functions!
    """
    x: float
    y: float
    z: float


@dataclass
class Rotator:
    """
    Values are in radians. For the mathematically inclined:
    Order of application is yaw, pitch, roll
    """
    pitch: float
    yaw: float
    roll: float


@dataclass
class Physics:
    location: Vector3
    rotation: Rotator
    velocity: Vector3
    angular_velocity: Vector3


@dataclass
class Touch:
    player_name: str
    time_seconds: float
    hit_location: Vector3
    hit_normal: Vector3
    team: int
    player_index: int


@dataclass
class ScoreInfo:
    score: int
    goals: int
    own_goals: int
    assists: int
    saves: int
    shots: int
    demolitions: int


@dataclass
class BoxShape:
    length: float
    width: float
    height: float


@dataclass
class SphereShape:
    diameter: float


@dataclass
class CylinderShape:
    diameter: float
    height: float


@dataclass
class ShapeType(IntEnum):
    box = 0
    sphere = 1
    cylinder = 2


@dataclass
class CollisionShape:
    type: ShapeType
    box: BoxShape
    sphere: SphereShape
    cylinder: CylinderShape


@dataclass
class PlayerInfo:
    physics: Physics
    score_info: ScoreInfo
    is_demolished: bool
    has_wheel_contact: bool
    is_super_sonic: bool
    is_bot: bool
    jumped: bool
    double_jumped: bool
    name: str
    team: int
    boost: int
    hitbox: BoxShape
    hitbox_offset: Vector3
    spawn_id: int


@dataclass
class DropShotInfo:
    absorbed_force: float
    damage_index: int
    force_accum_recent: float


@dataclass
class BallInfo:
    physics: Physics
    latest_touch: Touch
    drop_shot_info: DropShotInfo
    collision_shape: CollisionShape


@dataclass
class BoostPadState:
    is_active: bool
    timer: float


@dataclass
class TileInfo:
    tile_state: int


@dataclass
class TeamInfo:
    team_index: int
    score: int


@dataclass
class GameInfo:
    seconds_elapsed: float
    game_time_remaining: float
    is_overtime: bool
    is_unlimited_time: bool
    is_round_active: bool
    is_kickoff_pause: bool
    is_match_ended: bool
    world_gravity_z: bool
    game_speed: float
    frame_num: int


@dataclass
class GameTickPacket:
    game_cars: List[PlayerInfo]
    num_cars: int
    game_boosts: List[BoostPadState]
    num_boost: int
    game_ball: BallInfo
    game_info: GameInfo
    # dropshot_tiles: List[TileInfo]  Hiding this for now because we don't support dropshot at the moment
    # num_tiles: int  Hiding this for now because we don't support dropshot at the moment
    teams: List[TeamInfo]
    num_teams: int


@dataclass
class BoostPad:
    location: Vector3
    is_full_boost: bool


@dataclass
class GoalInfo:
    team_num: int
    location: Vector3
    direction: Vector3
    width: float
    height: float


@dataclass
class FieldInfoPacket:
    boost_pads: List[BoostPad]
    num_boosts: int
    goals: List[GoalInfo]
    num_goals: int

class DropshotTileState(IntEnum):
    UNKNOWN = 0
    FILLED = 1
    DAMAGED = 2
    OPEN = 3



import os
import socket
import time

import psutil

from rlbot.agents.base_independent_agent import BaseIndependentAgent
from rlbot.botmanager.helper_process_request import HelperProcessRequest
from rlbot.utils.structures import game_interface


class BaseDotNetAgent(BaseIndependentAgent):

    def __init__(self, name, team, index):
        super().__init__(name, team, index)
        self.port = self.read_port_from_file()
        self.is_retired = False
        self.dotnet_executable_path = None

    def run_independently(self, terminate_request_event):

        while not terminate_request_event.is_set():
            message = f"add\n{self.name}\n{self.team}\n{self.index}\n{game_interface.get_dll_directory()}"
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.connect(("127.0.0.1", self.port))
                s.send(bytes(message, "ASCII"))
                s.close()
            except ConnectionRefusedError:
                self.logger.warn("Could not connect to server!")

            time.sleep(1)
        else:
            self.retire()

    def get_helper_process_request(self):
        if self.is_executable_configured():
            return HelperProcessRequest(python_file_path=None, key=__file__ + str(self.port),
                                        executable=self.dotnet_executable_path,
                                        current_working_directory=os.path.dirname(self.dotnet_executable_path))
        return None

    def is_executable_configured(self):
        return self.dotnet_executable_path is not None and os.path.isfile(self.dotnet_executable_path)

    def get_extra_pids(self):
        """
        Gets the list of process ids that should be marked as high priority.
        :return: A list of process ids that are used by this bot in addition to the ones inside the python process.
        """
        while not self.is_retired:
            for proc in psutil.process_iter():
                for conn in proc.connections():
                    if conn.laddr.port == self.port:
                        self.logger.debug(f".NET socket server for {self.name} appears to have pid {proc.pid}")
                        return [proc.pid]
            if self.is_executable_configured():
                return []
            time.sleep(1)
            if self.dotnet_executable_path is None:
                self.logger.info("Can't auto-start .NET executable because no executable is configured. "
                                 "Please start the .NET bot manually!")
            else:
                self.logger.info(f"Can't auto-start .NET executable because {self.dotnet_executable_path} "
                                 "is not found. Please start the .NET bot manually!")

    def retire(self):
        port = self.read_port_from_file()
        message = f"remove\n{self.index}"

        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("127.0.0.1", port))
            s.send(bytes(message, "ASCII"))
            s.close()
        except ConnectionRefusedError:
            self.logger.warn("Could not connect to server!")
        self.is_retired = True

    def read_port_from_file(self):
        try:
            location = self.get_port_file_path()

            with open(location, "r") as port_file:
                return int(port_file.readline().rstrip())

        except ValueError:
            self.logger.warn("Failed to parse port file!")
            raise

    def get_port_file_path(self):
        raise NotImplementedError



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class RespawnTimeOption(object):
    Three_Seconds = 0
    Two_Seconds = 1
    One_Seconds = 2
    Disable_Goal_Reset = 3




from signal import SIGTERM
from subprocess import Popen

import queue
import os

import rlbot
from rlbot.botmanager.agent_metadata import AgentMetadata
from rlbot.botmanager.bot_helper_process import BotHelperProcess
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.structures.game_interface import get_dll_directory

class SubprocessHivemind(BotHelperProcess):

    # Terminology:
    # hivemind - main process controlling the drones.
    # drone - a bot under the hivemind's control.

    # Path to the executable.
    exec_path = None

    def __init__(self, agent_metadata_queue, quit_event, options):
        super().__init__(agent_metadata_queue, quit_event, options)

        # Sets up the logger. The string is the name of your hivemind.
        # This is configured inside your config file under hivemind_name.
        self.logger = get_logger(options['name'])

        # Give a warning if exec_path was given.
        if 'exec_path' is None:
            raise NotImplementedError(
                "Please specify the exec_path in your subclass of SubprocessHivemind.")

        # drone_indices is a set of bot indices
        # which requested this hivemind with the same key.
        self.drone_indices = set()

    def try_receive_agent_metadata(self):
        """Adds all drones with the correct key to our set of running indices."""
        while not self.metadata_queue.empty():
            # Tries to get the next agent from the queue.
            single_agent_metadata: AgentMetadata = self.metadata_queue.get(
                timeout=1.0)
            # Adds the index to the drone_indices.
            self.drone_indices.add(single_agent_metadata.index)

    def start(self):
        """Starts the BotHelperProcess - Hivemind."""
        # Prints an activation message into the console.
        # This lets you know that the process is up and running.
        self.logger.debug("Hello, world!")

        # Collect drone indices that requested a helper process with our key.
        self.logger.info("Collecting drones; give me a moment.")
        self.try_receive_agent_metadata()
        self.logger.info("Starting subprocess!")

        if not os.path.isfile(self.exec_path): 
            raise FileNotFoundError(f"Could not find file: {self.exec_path}")

        # Starts the process.
        # The drone indices are all the numbers after --drone-indices.
        # example.exe --rlbot-version 1.35.5 --rlbot-dll-directory some/directory/dll/ --drone-indices 0,1,2,3,4,5
        process = Popen(
            [
                self.exec_path,
                '--rlbot-version',
                rlbot.__version__,
                '--rlbot-dll-directory',
                get_dll_directory(),
                '--drone-indices',
                ','.join([str(index) for index in self.drone_indices])
            ]
        )

        # Wait until we get a quit_event, and then terminate the process.
        self.quit_event.wait()
        process.send_signal(SIGTERM)
        process.wait()



import ctypes
import math
from enum import IntEnum

from rlbot.utils.structures.start_match_structures import MAX_NAME_LENGTH, MAX_PLAYERS
from rlbot.utils.structures.struct import Struct
from rlbot.utils.structures.utils import create_enum_object

MAX_BOOSTS = 50
MAX_TILES = 200
MAX_TEAMS = 2
MAX_GOALS = 200


class Vector3(Struct):
    _fields_ = [("x", ctypes.c_float),
                ("y", ctypes.c_float),
                ("z", ctypes.c_float)]


# Note: this is now expressed in radians (it used to be unreal rotation units).
class Rotator(Struct):
    _fields_ = [("pitch", ctypes.c_float),
                ("yaw", ctypes.c_float),
                ("roll", ctypes.c_float)]


class Physics(Struct):
    _fields_ = [("location", Vector3),
                ("rotation", Rotator),
                ("velocity", Vector3),
                ("angular_velocity", Vector3)]


class Touch(Struct):
    _fields_ = [("player_name", ctypes.c_wchar * MAX_NAME_LENGTH),
                ("time_seconds", ctypes.c_float),
                ("hit_location", Vector3),
                ("hit_normal", Vector3),
                ("team", ctypes.c_int),
                ("player_index", ctypes.c_int)]


class ScoreInfo(Struct):
    # Describes the points for a single player (see TeamInfo for team scores)
    _fields_ = [("score", ctypes.c_int),
                ("goals", ctypes.c_int),
                ("own_goals", ctypes.c_int),
                ("assists", ctypes.c_int),
                ("saves", ctypes.c_int),
                ("shots", ctypes.c_int),
                ("demolitions", ctypes.c_int)]


class BoxShape(Struct):
    _fields_ = [("length", ctypes.c_float),
                ("width", ctypes.c_float),
                ("height", ctypes.c_float)]


class SphereShape(Struct):
    _fields_ = [("diameter", ctypes.c_float)]


class CylinderShape(Struct):
    _fields_ = [("diameter", ctypes.c_float),
                ("height", ctypes.c_float)]


class ShapeType(IntEnum):
    box = 0
    sphere = 1
    cylinder = 2


class CollisionShape(Struct):
    _fields_ = [("type", ctypes.c_int),
                ("box", BoxShape),
                ("sphere", SphereShape),
                ("cylinder", CylinderShape)]


class PlayerInfo(Struct):
    _fields_ = [("physics", Physics),
                ("score_info", ScoreInfo),
                ("is_demolished", ctypes.c_bool),
                # True if your wheels are on the ground, the wall, or the ceiling. False if you're midair or turtling.
                ("has_wheel_contact", ctypes.c_bool),
                ("is_super_sonic", ctypes.c_bool),
                ("is_bot", ctypes.c_bool),
                # True if the player has jumped. Falling off the ceiling / driving off the goal post does not count.
                ("jumped", ctypes.c_bool),
                # True if player has double jumped. False does not mean you have a jump remaining, because the
                # aerial timer can run out, and that doesn't affect this flag.
                ("double_jumped", ctypes.c_bool),
                ("name", ctypes.c_wchar * MAX_NAME_LENGTH),
                ("team", ctypes.c_ubyte),
                ("boost", ctypes.c_int),
                ("hitbox", BoxShape),
                ("hitbox_offset", Vector3),
                ("spawn_id", ctypes.c_int)]


class DropShotInfo(Struct):
    _fields_ = [("absorbed_force", ctypes.c_float),
                ("damage_index", ctypes.c_int),
                ("force_accum_recent", ctypes.c_float)]


class BallInfo(Struct):
    _fields_ = [("physics", Physics),
                ("latest_touch", Touch),
                ("drop_shot_info", DropShotInfo),
                ("collision_shape", CollisionShape)]


class BoostPadState(Struct):
    _fields_ = [("is_active", ctypes.c_bool),
                ("timer", ctypes.c_float)]


class TileInfo(Struct):
    _fields_ = [("tile_state", ctypes.c_int)]  # see DropShotTileState


class TeamInfo(Struct):
    _fields_ = [("team_index", ctypes.c_int),
                ("score", ctypes.c_int)]


class GameInfo(Struct):
    _fields_ = [("seconds_elapsed", ctypes.c_float),
                ("game_time_remaining", ctypes.c_float),
                ("is_overtime", ctypes.c_bool),
                ("is_unlimited_time", ctypes.c_bool),
                # True when cars are allowed to move, and during the pause menu. False during replays.
                ("is_round_active", ctypes.c_bool),
                # Only false during a kickoff, when the car is allowed to move, and the ball has not been hit,
                # and the game clock has not started yet. If both players sit still, game clock will eventually
                # start and this will become true.
                ("is_kickoff_pause", ctypes.c_bool),
                # Turns true after final replay, the moment the 'winner' screen appears. Remains true during next match
                # countdown. Turns false again the moment the 'choose team' screen appears.
                ("is_match_ended", ctypes.c_bool),
                ("world_gravity_z", ctypes.c_float),
                # Game speed multiplier, 1.0 is regular game speed.
                ("game_speed", ctypes.c_float),
                # Number of physics frames which have elapsed in the game.
                # May increase by more than one across consecutive packets.
                ("frame_num", ctypes.c_int)]


# On the c++ side this struct has a long at the beginning for locking.
# This flag is removed from this struct so it isn't visible to users.


class GameTickPacket(Struct):
    _fields_ = [("game_cars", PlayerInfo * MAX_PLAYERS),
                ("num_cars", ctypes.c_int),
                ("game_boosts", BoostPadState * MAX_BOOSTS),
                ("num_boost", ctypes.c_int),
                ("game_ball", BallInfo),
                ("game_info", GameInfo),
                ("dropshot_tiles", TileInfo * MAX_TILES),
                ("num_tiles", ctypes.c_int),
                ("teams", TeamInfo * MAX_TEAMS),
                ("num_teams", ctypes.c_int)]


class BoostPad(Struct):
    _fields_ = [("location", Vector3),
                ("is_full_boost", ctypes.c_bool)]


class GoalInfo(Struct):
    _fields_ = [("team_num", ctypes.c_ubyte),
                ("location", Vector3),
                ("direction", Vector3),
                ("width", ctypes.c_float),
                ("height", ctypes.c_float)]


class FieldInfoPacket(Struct):
    _fields_ = [("boost_pads", BoostPad * MAX_BOOSTS),
                ("num_boosts", ctypes.c_int),
                ("goals", GoalInfo * MAX_GOALS),
                ("num_goals", ctypes.c_int)]


# Helps us return raw bytes from protobuf functions.
class ByteBuffer(Struct):
    _fields_ = [("ptr", ctypes.c_void_p),
                ("size", ctypes.c_int)]


# This negates all x and y values for balls and cars and rotates yaw 180 degrees.
def rotate_game_tick_packet_boost_omitted(game_tick_packet):
    # Negate all x,y values for ball
    game_tick_packet.game_ball.physics.location.x = \
        -1 * game_tick_packet.game_ball.physics.location.x
    game_tick_packet.game_ball.physics.location.y = \
        -1 * game_tick_packet.game_ball.physics.location.y
    game_tick_packet.game_ball.physics.velocity.x = \
        -1 * game_tick_packet.game_ball.physics.velocity.x
    game_tick_packet.game_ball.physics.velocity.y = \
        -1 * game_tick_packet.game_ball.physics.velocity.y
    # Angular velocity is stored on global axis so negating on x and y does make sense!
    game_tick_packet.game_ball.physics.angular_velocity.x = \
        -1 * game_tick_packet.game_ball.physics.angular_velocity.x
    game_tick_packet.game_ball.physics.angular_velocity.y = \
        -1 * game_tick_packet.game_ball.physics.angular_velocity.y

    # ball touch data
    game_tick_packet.game_ball.latest_touch.hit_location.x = \
        -1 * game_tick_packet.game_ball.latest_touch.hit_location.x
    game_tick_packet.game_ball.latest_touch.hit_location.y = \
        -1 * game_tick_packet.game_ball.latest_touch.hit_location.y
    game_tick_packet.game_ball.latest_touch.hit_normal.x = \
        -1 * game_tick_packet.game_ball.latest_touch.hit_normal.x
    game_tick_packet.game_ball.latest_touch.hit_normal.y = \
        -1 * game_tick_packet.game_ball.latest_touch.hit_normal.y

    # Rotate yaw 180 degrees is all that is necessary.
    ball_yaw = game_tick_packet.game_ball.physics.rotation.yaw
    game_tick_packet.game_ball.physics.rotation.yaw = ball_yaw \
                                                      + math.pi if ball_yaw < 0 else ball_yaw - math.pi

    for i in range(game_tick_packet.num_cars):
        game_tick_packet.game_cars[i].physics.location.x = \
            -1 * game_tick_packet.game_cars[i].physics.location.x
        game_tick_packet.game_cars[i].physics.location.y = \
            -1 * game_tick_packet.game_cars[i].physics.location.y
        game_tick_packet.game_cars[i].physics.velocity.x = \
            -1 * game_tick_packet.game_cars[i].physics.velocity.x
        game_tick_packet.game_cars[i].physics.velocity.y = \
            -1 * game_tick_packet.game_cars[i].physics.velocity.y
        game_tick_packet.game_cars[i].physics.angular_velocity.x = \
            -1 * game_tick_packet.game_cars[i].physics.angular_velocity.x
        game_tick_packet.game_cars[i].physics.angular_velocity.y = \
            -1 * game_tick_packet.game_cars[i].physics.angular_velocity.y

        car_yaw = game_tick_packet.game_cars[i].physics.rotation.yaw
        game_tick_packet.game_cars[i].physics.rotation.yaw = car_yaw \
                                                             + math.pi if car_yaw < 0 else car_yaw - math.pi


DropshotTileState = create_enum_object(["UNKNOWN",
                                        "FILLED",
                                        "DAMAGED",
                                        "OPEN"])



import ctypes
import hashlib
from typing import Optional, Set

from rlbot.utils.structures.game_data_struct import Vector3
from rlbot.utils.structures.struct import Struct
from rlbot.utils.logging_utils import get_logger

class Color(Struct):
    _fields_ = [("r", ctypes.c_ubyte),
                ("g", ctypes.c_ubyte),
                ("b", ctypes.c_ubyte),
                ("a", ctypes.c_ubyte)]


MAX_INT = 2147483647 // 2

DEFAULT_GROUP_ID = 'default'

class RenderingManager:
    def __init__(self):
        self.renderGroup = None
        self.render_state = False
        self.builder = None
        self.render_list = []
        self.bot_index = 0
        self.bot_team = 0
        self.group_id: Optional[str] = None
        self.touched_group_ids: Set[str] = set()

        self.native_constructor = None
        self.native_destructor = None
        self.native_finish_and_send = None
        self.native_draw_line_3d = None
        self.native_draw_polyline_2d = None
        self.native_draw_string_2d = None
        self.native_draw_string_3d = None
        self.native_draw_rect_2d = None
        self.native_draw_rect_3d = None

    def setup_function_types(self, dll_instance):
        self.native_constructor = dll_instance.Renderer_Constructor
        self.native_constructor.argtypes = [ctypes.c_int]
        self.native_constructor.restype = ctypes.c_void_p

        self.native_destructor = dll_instance.Renderer_Destructor
        self.native_destructor.argtypes = [ctypes.c_void_p]

        self.native_finish_and_send = dll_instance.Renderer_FinishAndSend
        self.native_finish_and_send.argtypes = [ctypes.c_void_p]

        self.native_draw_line_3d = dll_instance.Renderer_DrawLine3D
        self.native_draw_line_3d.argtypes = [ctypes.c_void_p, Color, Vector3, Vector3]

        self.native_draw_polyline_3d = dll_instance.Renderer_DrawPolyLine3D
        self.native_draw_polyline_3d.argtypes = [ctypes.c_void_p, Color, ctypes.POINTER(Vector3), ctypes.c_int]

        self.native_draw_string_2d = dll_instance.Renderer_DrawString2D
        self.native_draw_string_2d.argtypes = [ctypes.c_void_p, ctypes.c_char_p, Color, Vector3, ctypes.c_int, ctypes.c_int]

        self.native_draw_string_3d = dll_instance.Renderer_DrawString3D
        self.native_draw_string_3d.argtypes = [ctypes.c_void_p, ctypes.c_char_p, Color, Vector3, ctypes.c_int, ctypes.c_int]

        self.native_draw_rect_2d = dll_instance.Renderer_DrawRect2D
        self.native_draw_rect_2d.argtypes = [ctypes.c_void_p, Color, Vector3, ctypes.c_int, ctypes.c_int, ctypes.c_bool]

        self.native_draw_rect_3d = dll_instance.Renderer_DrawRect3D
        self.native_draw_rect_3d.argtypes = [ctypes.c_void_p, Color, Vector3, ctypes.c_int, ctypes.c_int, ctypes.c_bool, ctypes.c_bool]

    def set_bot_index_and_team(self, bot_index=0, bot_team=0):
        self.bot_index = bot_index
        self.bot_team = bot_team

    def send_group(self, buffer):
        rlbot_status = self.renderGroup(bytes(buffer), len(buffer))
        if rlbot_status != RLBotCoreStatus.Success:
            get_logger("Renderer").error("bad status %s", RLBotCoreStatus.status_list[rlbot_status])

    def begin_rendering(self, group_id: str=DEFAULT_GROUP_ID):
        self.touched_group_ids.add(group_id)
        self.group_id = group_id
        self.render_list = []
        self.render_state = True

        if self.group_id is None:
            self.group_id = DEFAULT_GROUP_ID

        group_id = str(self.bot_index) + str(self.group_id)
        group_id_hashed = int(hashlib.sha256(str(group_id).encode('utf-8')).hexdigest(), 16) % MAX_INT

        # Prevent memory leak
        if self.builder is not None:
            self.native_destructor(self.builder)

        self.builder = self.native_constructor(group_id_hashed)


    def end_rendering(self):
        if self.builder is not None:
            self.native_finish_and_send(self.builder)
            self.native_destructor(self.builder)
        self.builder = None
        
        self.render_state = False

    def clear_screen(self, group_id: str=DEFAULT_GROUP_ID):
        self.begin_rendering(group_id)
        self.end_rendering()

    def clear_all_touched_render_groups(self):
        """
        Clears all render groups which have been drawn to using `begin_rendering(group_id)`.
        Note: This does not clear render groups created by e.g. other bots.
        """
        for group_id in self.touched_group_ids:
            self.clear_screen(group_id)

    def is_rendering(self):
        return self.render_state

    def draw_line_2d(self, x1, y1, x2, y2, color):
        get_logger("Renderer").error("draw_line_2d is not supported!")
        return self

    def draw_polyline_2d(self, vectors, color):
        if len(vectors) < 2:
            get_logger("Renderer").error("draw_polyline_2d requires atleast 2 vectors!")
            return self

        get_logger("Renderer").error("draw_polyline_2d is not supported!")
        return self

    def draw_line_3d(self, vec1, vec2, color):
        if self.builder is None:
            get_logger("Renderer").error("Use begin_rendering before using any of the drawing functions!")
            return self

        self.native_draw_line_3d(self.builder, color, self.__create_vector(vec1), self.__create_vector(vec2))
        return self

    def draw_polyline_3d(self, vectors, color):
        if len(vectors) < 2:
            get_logger("Renderer").error("draw_polyline_3d requires atleast 2 vectors!")
            return self

        if self.builder is None:
            get_logger("Renderer").error("Use begin_rendering before using any of the drawing functions!")
            return self
        
        # TODO: would be nice to use a native function
        for i in range(len(vectors) - 1):
            self.draw_line_3d(vectors[i], vectors[i+1], color)

        return self

    def draw_line_2d_3d(self, x, y, vec, color):
        get_logger("Renderer").error("draw_line_2d_3d is not supported!")
        return self

    def draw_rect_2d(self, x, y, width, height, filled, color):
        if self.builder is None:
            get_logger("Renderer").error("Use begin_rendering before using any of the drawing functions!")
            return self
        self.native_draw_rect_2d(self.builder, color, self.__create_vector([x,y,0]), width, height, filled)
        return self

    def draw_rect_3d(self, vec, width, height, filled, color, centered=False):
        if self.builder is None:
            get_logger("Renderer").error("Use begin_rendering before using any of the drawing functions!")
            return self
        self.native_draw_rect_3d(self.builder, color, self.__create_vector(vec), width, height, filled, centered)
        return self

    def draw_string_2d(self, x, y, scale_x, scale_y, text, color):
        if self.builder is None:
            get_logger("Renderer").error("Use begin_rendering before using any of the drawing functions!")
            return self
        self.native_draw_string_2d(self.builder, text.encode('utf-8'), color, self.__create_vector([x,y,0]), scale_x, scale_y)
        return self

    def draw_string_3d(self, vec, scale_x, scale_y, text, color):
        if self.builder is None:
            get_logger("Renderer").error("Use begin_rendering before using any of the drawing functions!")
            return self
        self.native_draw_string_3d(self.builder, text.encode('utf-8'), color, self.__create_vector(vec), scale_x, scale_y)
        return self

    def create_color(self, alpha, red, green, blue):
        return Color(red, green, blue, alpha)

    def black(self):
        return self.create_color(255, 0, 0, 0)

    def white(self):
        return self.create_color(255, 255, 255, 255)

    def gray(self):
        return self.create_color(255, 128, 128, 128)

    def grey(self):
        return self.gray()

    def blue(self):
        return self.create_color(255, 0, 0, 255)

    def red(self):
        return self.create_color(255, 255, 0, 0)

    def green(self):
        return self.create_color(255, 0, 128, 0)

    def lime(self):
        return self.create_color(255, 0, 255, 0)

    def yellow(self):
        return self.create_color(255, 255, 255, 0)

    def orange(self):
        return self.create_color(255, 225, 128, 0)

    def cyan(self):
        return self.create_color(255, 0, 255, 255)

    def pink(self):
        return self.create_color(255, 255, 0, 255)

    def purple(self):
        return self.create_color(255, 128, 0, 128)

    def teal(self):
        return self.create_color(255, 0, 128, 128)

    def team_color(self, team=None, alt_color=False):
        """
        Returns the team color of the bot. Team 0: blue, team 1: orange, other: white
        :param team: Specify which team's color. If None, the associated bot's team color will be returned.
        :param alt_color: If True, returns the alternate team colors instead. Team 0: cyan, team 1: red, other: gray
        :return: a team color
        """
        if team is None:
            team = self.bot_team

        if team == 0:
            return self.cyan() if alt_color else self.blue()
        elif team == 1:
            return self.red() if alt_color else self.orange()
        else:
            return self.gray() if alt_color else self.white()

    def get_rendering_manager(self, bot_index=0, bot_team=0):
        """
        Gets all the raw render functions but without giving access to any internal logic or the dll
        :return: An object with the same interface as the functions above
        """
        self.set_bot_index_and_team(bot_index, bot_team)
        return self

    def __wrap_float(self, number):
        return Float.CreateFloat(self.builder, number)

    def __create_vector(self, *vec) -> Vector3:
        """
        Converts a variety of vector types to a flatbuffer Vector3.
        Supports Flatbuffers Vector3, cTypes Vector3, list/tuple of numbers, or passing x,y,z (z optional)
        """
        import numbers

        if len(vec) == 1:
            if hasattr(vec[0], "__getitem__"):  # Support all subscriptable types.
                try:
                    x = float(vec[0][0])
                    y = float(vec[0][1])
                    try:
                        z = float(vec[0][2])
                    except (ValueError, IndexError):
                        z = 0
                except ValueError:
                    raise ValueError(f"Unexpected type(s) for creating vector: {type(vec[0][0])}, {type(vec[0][1])}")
                except IndexError:
                    raise IndexError(f"Unexpected IndexError when creating vector from type: {type(vec[0])}")
            #elif isinstance(vec[0], Vector3.Vector3):
            #    x = vec[0].X()
            #    y = vec[0].Y()
            #    z = vec[0].Z()
            elif isinstance(vec[0], Vector3):
                x = vec[0].x
                y = vec[0].y
                z = vec[0].z
            else:
                raise ValueError(f"Unexpected type for creating vector: {type(vec[0])}")
        elif len(vec) == 2 or len(vec) == 3:
            if isinstance(vec[0], numbers.Number) and isinstance(vec[1], numbers.Number):
                x = vec[0]
                y = vec[1]
                if len(vec) == 2:
                    z = 0
                else:
                    if isinstance(vec[2], numbers.Number):
                        z = vec[2]
                    else:
                        raise ValueError(f"Unexpected type for creating vector: {type(vec[0])}")
            else:
                raise ValueError(f"Unexpected type(s) for creating vector: {type(vec[0])}, {type(vec[1])}")
        else:
            raise ValueError("Unexpected number of arguments for creating vector")

        return Vector3(x, y, z)


class DummyRenderer(RenderingManager):

    def __init__(self, renderer):
        self.renderGroup = renderer.renderGroup
        self.render_state = renderer.render_state
        self.builder = renderer.builder
        self.render_list = renderer.render_list
        self.group_id = renderer.group_id
        self.bot_index = renderer.bot_index
        self.bot_team = renderer.bot_team

    def clear_screen(self, group_id='default'):
        pass

    def draw_line_2d(self, x1, y1, x2, y2, color):
        return self

    def draw_polyline_2d(self, vectors, color):
        return self

    def draw_line_3d(self, vec1, vec2, color):
        return self

    def draw_polyline_3d(self, vectors, color):
        return self

    def draw_line_2d_3d(self, x, y, vec, color):
        return self

    def draw_rect_2d(self, x, y, width, height, filled, color):
        return self

    def draw_rect_3d(self, vec, width, height, filled, color, centered=False):
        return self

    def draw_string_2d(self, x, y, scale_x, scale_y, text, color):
        return self

    def draw_string_3d(self, vec, scale_x, scale_y, text, color):
        return self

    def begin_rendering(self, group_id='default'):
        self.builder = flatbuffers.Builder(0)

    def end_rendering(self):
        pass

    def clear_all_touched_render_groups(self):
        pass



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DesiredGameInfoState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDesiredGameInfoState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DesiredGameInfoState()
        x.Init(buf, n + offset)
        return x

    # DesiredGameInfoState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DesiredGameInfoState
    def WorldGravityZ(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameInfoState
    def GameSpeed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameInfoState
    def Paused(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Bool import Bool
            obj = Bool()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DesiredGameInfoState
    def EndMatch(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Bool import Bool
            obj = Bool()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def DesiredGameInfoStateStart(builder): builder.StartObject(4)
def DesiredGameInfoStateAddWorldGravityZ(builder, worldGravityZ): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(worldGravityZ), 0)
def DesiredGameInfoStateAddGameSpeed(builder, gameSpeed): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(gameSpeed), 0)
def DesiredGameInfoStateAddPaused(builder, paused): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(paused), 0)
def DesiredGameInfoStateAddEndMatch(builder, endMatch): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(endMatch), 0)
def DesiredGameInfoStateEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class FieldInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsFieldInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = FieldInfo()
        x.Init(buf, n + offset)
        return x

    # FieldInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # FieldInfo
    def BoostPads(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .BoostPad import BoostPad
            obj = BoostPad()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # FieldInfo
    def BoostPadsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # FieldInfo
    def Goals(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .GoalInfo import GoalInfo
            obj = GoalInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # FieldInfo
    def GoalsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def FieldInfoStart(builder): builder.StartObject(2)
def FieldInfoAddBoostPads(builder, boostPads): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(boostPads), 0)
def FieldInfoStartBoostPadsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def FieldInfoAddGoals(builder, goals): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(goals), 0)
def FieldInfoStartGoalsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def FieldInfoEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Rocket League is notifying us that some player has moved their controller. This is an *output*
class PlayerInputChange(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayerInputChange(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PlayerInputChange()
        x.Init(buf, n + offset)
        return x

    # PlayerInputChange
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PlayerInputChange
    def PlayerIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # PlayerInputChange
    def ControllerState(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .ControllerState import ControllerState
            obj = ControllerState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # PlayerInputChange
    def DodgeForward(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # PlayerInputChange
    def DodgeRight(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def PlayerInputChangeStart(builder): builder.StartObject(4)
def PlayerInputChangeAddPlayerIndex(builder, playerIndex): builder.PrependInt32Slot(0, playerIndex, 0)
def PlayerInputChangeAddControllerState(builder, controllerState): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(controllerState), 0)
def PlayerInputChangeAddDodgeForward(builder, dodgeForward): builder.PrependFloat32Slot(2, dodgeForward, 0.0)
def PlayerInputChangeAddDodgeRight(builder, dodgeRight): builder.PrependFloat32Slot(3, dodgeRight, 0.0)
def PlayerInputChangeEnd(builder): return builder.EndObject()



from inputs import get_gamepad  # pip install inputs
from rlbot.agents.base_agent import SimpleControllerState
import threading

def deadzone(normalized_axis):
    if abs(normalized_axis) < 0.1: return 0.0
    return normalized_axis

class ControllerInput:
    def __init__(self):
        self._gas_pedal = 0.0
        self._brake_pedal = 0.0

        self._left_stick_x = 0.0
        self._left_stick_y = 0.0

        self._right_stick_x = 0.0
        self._right_stick_y = 0.0

        self._north = 0
        self._east = 0
        self._south = 0
        self._west = 0

        self._hat_x = 0
        self._hat_y = 0

        self._tl = 0
        self._tr = 0

        self._thumb_l = 0
        self._thumb_r = 0

        self._select = 0
        self._start = 0


    @property
    def throttle(self):
        return self._gas_pedal - self._brake_pedal

    @property
    def steer(self):
        return self._left_stick_x

    @property
    def roll(self):
        return self._left_stick_x if self.aerial_control else 0.0

    @property
    def yaw(self):
        return 0.0 if self.aerial_control else self._left_stick_x

    @property
    def pitch(self):
        return self._left_stick_y

    @property
    def jump(self):
        return self._south

    @property
    def boost(self):
        return self._east

    @property
    def handbrake(self):
        return self._west 

    @property
    def aerial_control(self):
        return self._tr

    def get_output(self):
        output = SimpleControllerState()

        output.throttle = self.throttle
        output.steer = self.steer
        output.pitch = self.pitch
        output.yaw = self.yaw
        output.roll = self.roll
        output.jump = self.jump
        output.boost = self.boost
        output.handbrake = self.handbrake

        return output

    def main_poll_loop(self):
        while 1:
            events = get_gamepad()  # Blocking
            for event in events:
                #print(repr((event.ev_type, event.code, event.state)))
                if False: pass
                elif event.ev_type=='Absolute' and event.code=='ABS_RZ': self._gas_pedal = deadzone(event.state / 255.0)
                elif event.ev_type=='Absolute' and event.code=='ABS_Z': self._brake_pedal = deadzone(event.state / 255.0)
                elif event.ev_type=='Absolute' and event.code=='ABS_X': self._left_stick_x = deadzone(event.state / 32768.0)
                elif event.ev_type=='Absolute' and event.code=='ABS_Y': self._left_stick_y = deadzone(-event.state / 32768.0)
                elif event.ev_type=='Absolute' and event.code=='ABS_RX': self._right_stick_x = deadzone(event.state / 32768.0)
                elif event.ev_type=='Absolute' and event.code=='ABS_RY': self._right_stick_y = deadzone(-event.state / 32768.0)
                elif event.ev_type=='Absolute' and event.code=='ABS_HAT0X': self._hat_x = event.state
                elif event.ev_type=='Absolute' and event.code=='ABS_HAT0Y': self._hat_y = event.state
                elif event.ev_type=='Key' and event.code=='BTN_NORTH': self._north = event.state
                elif event.ev_type=='Key' and event.code=='BTN_EAST': self._east = event.state
                elif event.ev_type=='Key' and event.code=='BTN_SOUTH': self._south = event.state
                elif event.ev_type=='Key' and event.code=='BTN_WEST': self._west = event.state
                elif event.ev_type=='Key' and event.code=='BTN_TL': self._tl = event.state
                elif event.ev_type=='Key' and event.code=='BTN_TR': self._tr = event.state
                elif event.ev_type=='Key' and event.code=='BTN_THUMBL': self._thumb_l = event.state
                elif event.ev_type=='Key' and event.code=='BTN_THUMBR': self._thumb_r = event.state
                elif event.ev_type=='Key' and event.code=='BTN_SELECT': self._select = event.state
                elif event.ev_type=='Key' and event.code=='BTN_START': self._start = event.state
                elif event.ev_type!='Sync': print(repr((event.ev_type, event.code, event.state)))

                
    
    @property
    def left_stick_x(self):
        return self._left_stick_x

    @property
    def left_stick_y(self):
        return self._left_stick_y

    @property
    def right_stick_x(self):
        return self._right_stick_x

    @property
    def right_stick_y(self):
        return self._right_stick_y

    @property
    def up(self):
        return True if self._hat_y==-1 else False

    @property  
    def down(self):
        return True if self._hat_y==1 else False

    @property
    def left(self):
        return True if self._hat_x==-1 else False

    @property
    def right(self):
        return True if self._hat_x==1 else False

    @property
    def start(self):
        return self._select


    #xbox exclusive
    @property
    def A(self):
        return self._south

    @property
    def B(self):
        return self._east
    
    @property
    def X(self):
        return self._west
    
    @property
    def Y(self):
        return self._north

    @property
    def left_back(self):
        return self._tl
    
    @property
    def right_back(self):
        return self._tr

    @property
    def left_trigger(self):
        return self._brake_pedal
    
    @property
    def right_trigger(self):
        return self._gas_pedal

    @property
    def left_thumb(self):
        return self._thumb_l

    @property
    def right_thumb(self):
        return self._thumb_r
    
    @property
    def back(self):
        return self._start


    #ps3 exclusive
    @property
    def cross(self):
        return self._south

    @property
    def circle(self):
        return self._east
    
    @property
    def square(self):
        return self._west
    
    @property
    def triangle(self):
        return self._north

    @property
    def L1(self):
        return self._tl
    
    @property
    def R1(self):
        return self._tr

    @property
    def L2(self):
        return self._brake_pedal
    
    @property
    def R2(self):
        return self._gas_pedal

    @property
    def L3(self):
        return self._thumb_l

    @property
    def R3(self):
        return self._thumb_r
    
    @property
    def select(self):
        return self._start
                

controller = ControllerInput()
threading.Thread(target=controller.main_poll_loop, daemon=True).start()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class RenderType(object):
    DrawLine2D = 1
    DrawLine3D = 2
    DrawLine2D_3D = 3
    DrawRect2D = 4
    DrawRect3D = 5
    DrawString2D = 6
    DrawString3D = 7
    DrawCenteredRect3D = 8




import inspect
import multiprocessing as mp
import sys
from pathlib import Path
from typing import Type

from rlbot.agents.standalone.test_spawner import TestSpawner
from rlbot.agents.base_agent import BaseAgent
from rlbot.agents.standalone.standalone_bot_config import StandaloneBotConfig
from rlbot.botmanager.bot_manager_struct import BotManagerStruct
from rlbot.parsing.bot_config_bundle import get_bot_config_bundle
from rlbot.utils.class_importer import ExternalClassWrapper


class StandaloneBot(BaseAgent):
    """
    This is intended to be a base class for python bots. Your bot file would be something like this:

    from rlbot.agents.bot_manager_standalone_agent import BotManagerStandaloneAgent, run_bot

    class MyCoolBot(BotManagerStandaloneAgent):
        def get_output(self, packet: GameTickPacket):
            ...

    if __name__ == '__main__':
        run_bot(MyCoolBot)


    And also you would need your config file to have supports_standalone = True

    """

    def __init__(self, name, team, index):
        super().__init__(name, team, index)


def run_bot(agent_class: Type[StandaloneBot]):
    config = StandaloneBotConfig(sys.argv)
    python_file = inspect.getfile(agent_class)

    config_obj = agent_class.base_create_agent_configurations()
    bundle = None
    if config.config_file is not None:
        # If the config file was not passed, then the bot will miss out on any custom configuration,
        # tick rate preference, etc.
        bundle = get_bot_config_bundle(Path(python_file).parent / config.config_file)
        config_obj.parse_file(bundle.config_obj, config_directory=bundle.config_directory)

    spawn_id = config.spawn_id
    player_index = config.player_index
    team = config.team
    name = config.name

    if config.is_missing_args:
        # This process must not have been created by the RLBot framework, so this is probably
        # a developer doing some testing who did not pass all the params. Take it upon ourselves
        # to fire up the game if necessary.
        print(f'############################################################################################')
        print(f'Args are missing, so we will assume this is a dev workflow and insert the bot into the game!')
        print(f'############################################################################################')
        test_spawner = TestSpawner(Path(python_file), config, bundle)
        test_spawner.spawn_bot()
        spawn_id = test_spawner.spawn_id
        player_index = test_spawner.player_index
        team = test_spawner.team
        name = test_spawner.name


    agent_class_wrapper = ExternalClassWrapper(python_file, StandaloneBot)

    # Pass in dummy objects for mp.Event, mp.Queue. We will not support that
    # functionality for standalone bots; it's generally unused anyway.
    bot_manger = BotManagerStruct(terminate_request_event=mp.Event(),
                                  termination_complete_event=mp.Event(),
                                  reload_request_event=mp.Event(),
                                  bot_configuration=config_obj,
                                  name=name,
                                  team=team,
                                  index=player_index,
                                  agent_class_wrapper=agent_class_wrapper,
                                  agent_metadata_queue=mp.Queue(),
                                  match_config=None,
                                  matchcomms_root=config.matchcomms_url,
                                  spawn_id=spawn_id)
    bot_manger.run()



from typing import Type


class RLBotException(Exception):
    """Base class for exceptions in this module."""

    def __init__(self, msg=None):
        if msg is None:
            # Set some default useful error message
            msg = "An error occurred attempting to set RLBot configuration on the dll side"
        super(RLBotException, self).__init__(msg)


class BufferOverfilledError(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Buffer overfilled")


class MessageLargerThanMaxError(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Message larger than max")


class InvalidNumPlayerError(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid number of cars specified in configuration")


class InvalidBotSkillError(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid bot skill specified in configuration")


class InvalidHumanIndex(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid human index")


class InvalidPlayerIndexError(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid player index")


class InvalidName(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid name error. Python side was supposed to properly sanitize this!")


class InvalidTeam(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid team specified in configuration")


class InvalidTeamColor(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid team color specified")


class InvalidCustomColor(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid custom color specified")


class InvalidGameValues(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid game values")


class InvalidThrottle(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid throttle input")


class InvalidSteer(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid steer input")


class InvalidPitch(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid pitch input")


class InvalidYaw(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid yaw input")


class InvalidRoll(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid roll input")


class InvalidQuickChatPreset(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid quick chat preset")


class InvalidRenderType(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Invalid render type")


class EmptyDllResponse(RLBotException):
    def __init__(self):
        super(RLBotException, self).__init__("Response from dll was empty")


error_dict = {1: BufferOverfilledError, 2: MessageLargerThanMaxError, 3: InvalidNumPlayerError,
              4: InvalidBotSkillError, 5: InvalidHumanIndex, 6: InvalidName,
              7: InvalidTeam, 8: InvalidTeamColor, 9: InvalidCustomColor, 10: InvalidGameValues,
              11: InvalidThrottle, 12: InvalidSteer, 13: InvalidPitch, 14: InvalidYaw, 15: InvalidRoll,
              16: InvalidPlayerIndexError, 17: InvalidQuickChatPreset, 18: InvalidRenderType}


# https://stackoverflow.com/questions/33533148/how-do-i-specify-that-the-return-type-of-a-method-is-the-same-as-the-class-itsel
# https://docs.python.org/3/library/typing.html#classes-functions-and-decorators
def get_exception_from_error_code(error_code) -> Type['RLBotException']:
    try:
        return error_dict[error_code]
    except KeyError:
        return RLBotException



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class GameMessage(object):
    NONE = 0
    PlayerStatEvent = 1
    PlayerSpectate = 2
    PlayerInputChange = 3




from rlbot.setup_manager import SetupManager
from rlbot.utils.python_version_check import check_python_version


def main():

    print("starting")
    check_python_version()
    manager = SetupManager()
    manager.load_config()
    manager.connect_to_game()
    manager.launch_early_start_bot_processes()
    manager.start_match()
    manager.launch_bot_processes()
    manager.infinite_loop()  # Runs forever until interrupted


if __name__ == '__main__':
    main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DesiredBallState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDesiredBallState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DesiredBallState()
        x.Init(buf, n + offset)
        return x

    # DesiredBallState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DesiredBallState
    def Physics(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .DesiredPhysics import DesiredPhysics
            obj = DesiredPhysics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def DesiredBallStateStart(builder): builder.StartObject(1)
def DesiredBallStateAddPhysics(builder, physics): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(physics), 0)
def DesiredBallStateEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class DemolishOption(object):
    Default = 0
    Disabled = 1
    Friendly_Fire = 2
    On_Contact = 3
    On_Contact_FF = 4




import ctypes

MAX_PLAYERS = 64
MAX_NAME_LENGTH = 32


def get_player_configuration_list_type():
    return PlayerConfiguration * MAX_PLAYERS


class Color(ctypes.Structure):
    _fields_ = [("r", ctypes.c_ubyte),
                ("g", ctypes.c_ubyte),
                ("b", ctypes.c_ubyte),
                ("a", ctypes.c_ubyte)]


class PlayerConfiguration(ctypes.Structure):
    _fields_ = [("bot", ctypes.c_bool),
                ("rlbot_controlled", ctypes.c_bool),
                ("bot_skill", ctypes.c_float),
                ("human_index", ctypes.c_int),
                ("name", ctypes.c_wchar * MAX_NAME_LENGTH),
                ("team", ctypes.c_ubyte),
                ("team_color_id", ctypes.c_ubyte),
                ("custom_color_id", ctypes.c_ubyte),
                ("car_id", ctypes.c_int),
                ("decal_id", ctypes.c_int),
                ("wheels_id", ctypes.c_int),
                ("boost_id", ctypes.c_int),
                ("antenna_id", ctypes.c_int),
                ("hat_id", ctypes.c_int),
                ("paint_finish_id", ctypes.c_int),
                ("custom_finish_id", ctypes.c_int),
                ("engine_audio_id", ctypes.c_int),
                ("trails_id", ctypes.c_int),
                ("goal_explosion_id", ctypes.c_int),
                ("car_paint_id", ctypes.c_int),
                ("decal_paint_id", ctypes.c_int),
                ("wheels_paint_id", ctypes.c_int),
                ("boost_paint_id", ctypes.c_int),
                ("antenna_paint_id", ctypes.c_int),
                ("hat_paint_id", ctypes.c_int),
                ("trails_paint_id", ctypes.c_int),
                ("goal_explosion_paint_id", ctypes.c_int),
                ("use_rgb_lookup", ctypes.c_bool),
                ("primary_color_lookup", Color),
                ("secondary_color_lookup", Color),
                ("spawn_id", ctypes.c_int)]


class MutatorSettings(ctypes.Structure):
    _fields_ = [("match_length", ctypes.c_uint),
                ("max_score", ctypes.c_uint),
                ("overtime_option", ctypes.c_uint),
                ("series_length_option", ctypes.c_uint),
                ("game_speed_option", ctypes.c_uint),
                ("ball_max_speed_option", ctypes.c_uint),
                ("ball_type_option", ctypes.c_uint),
                ("ball_weight_option", ctypes.c_uint),
                ("ball_size_option", ctypes.c_uint),
                ("ball_bounciness_option", ctypes.c_uint),
                ("boost_amount_option", ctypes.c_uint),
                ("rumble_option", ctypes.c_uint),
                ("boost_strength_option", ctypes.c_uint),
                ("gravity_option", ctypes.c_uint),
                ("demolish_option", ctypes.c_uint),
                ("respawn_time_option", ctypes.c_uint),
                ]


class MatchSettings(ctypes.Structure):
    _fields_ = [("player_configuration", get_player_configuration_list_type()),
                ("num_players", ctypes.c_int),
                ("game_mode", ctypes.c_uint),
                ("game_map", ctypes.c_uint),
                ("skip_replays", ctypes.c_bool),
                ("instant_start", ctypes.c_bool),
                ("mutator_settings", MutatorSettings),
                ("existing_match_behavior", ctypes.c_uint),
                ("enable_lockstep", ctypes.c_bool),
                ("enable_rendering", ctypes.c_bool),
                ("enable_state_setting", ctypes.c_bool),
                ("auto_save_replay", ctypes.c_bool),
                ]



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class DesiredBoostState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDesiredBoostState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DesiredBoostState()
        x.Init(buf, n + offset)
        return x

    # DesiredBoostState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DesiredBoostState
    def RespawnTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Float import Float
            obj = Float()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def DesiredBoostStateStart(builder): builder.StartObject(1)
def DesiredBoostStateAddRespawnTime(builder, respawnTime): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(respawnTime), 0)
def DesiredBoostStateEnd(builder): return builder.EndObject()



import sys
from pathlib import Path
from subprocess import PIPE, Popen, run
from threading import Thread
from types import SimpleNamespace
from typing import List
from venv import EnvBuilder

from rlbot.parsing.bot_config_bundle import RunnableConfigBundle


class EnvBuilderWithRequirements(EnvBuilder):

    def __init__(self, bundle: RunnableConfigBundle):
        super().__init__(system_site_packages=True, clear=False, with_pip=False)
        self.bundle = bundle

    def post_setup(self, context: SimpleNamespace) -> None:
        requirements = self.bundle.requirements_file
        if not requirements:
            raise ValueError(f'Requirements file was not specified in {self.bundle.config_path}!')
        elif not Path(requirements).exists():
            raise ValueError(f'Requirements file {requirements} was not found!')
        sys.stderr.write(f'Installing {requirements}...\n')
        sys.stderr.flush()

        args = [context.env_exe, '-m', 'ensurepip']
        finished_process = self.run_and_dump(args, timeout=120)

        # Install in the virtual environment
        args = [context.env_exe, '-m', 'pip', 'install', '-U', '-r', requirements]
        finished_process = self.run_and_dump(args, timeout=300)

        if finished_process.returncode > 0:
            sys.stderr.write('FAILED to install requirements!')
            return
        sys.stderr.write('done.\n')

    def run_and_dump(self, args: List[str], timeout: int):
        finished_process = run(args, cwd=self.bundle.config_directory, capture_output=False, timeout=timeout)
        return finished_process


def setup_virtual_environment(runnable: RunnableConfigBundle):
    if not runnable.use_virtual_environment or not runnable.requirements_file:
        raise ValueError(f'{runnable.name} is not configured for virtual environment support!')
    builder = EnvBuilderWithRequirements(bundle=runnable)
    builder.create(Path(runnable.config_directory) / 'venv')




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class GoalInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGoalInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GoalInfo()
        x.Init(buf, n + offset)
        return x

    # GoalInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GoalInfo
    def TeamNum(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # GoalInfo
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GoalInfo
    def Direction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GoalInfo
    def Width(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # GoalInfo
    def Height(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def GoalInfoStart(builder): builder.StartObject(5)
def GoalInfoAddTeamNum(builder, teamNum): builder.PrependInt32Slot(0, teamNum, 0)
def GoalInfoAddLocation(builder, location): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def GoalInfoAddDirection(builder, direction): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(direction), 0)
def GoalInfoAddWidth(builder, width): builder.PrependFloat32Slot(3, width, 0.0)
def GoalInfoAddHeight(builder, height): builder.PrependFloat32Slot(4, height, 0.0)
def GoalInfoEnd(builder): return builder.EndObject()
